#
# Copyright (c) 2013,2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""This module contains the functionality used to setup relication between
the masters of Groups. The module contains utility methods for doing the
following

- Setting up replication between two groups - setup_group_replication
- Starting the slaves of a particular group - start_group_slaves
- Stopping the slaves of a particular group - stop_group_slaves
- Stopping a given group that is a slave - stop_group_slave

Note that each group is aware of the master group it replicates from and also
of the slave groups that replicate from it. This information helps in performing
clean up during the event of a "master changing operation" in the group.

As a design alternative these methods could have have been moved into the
mysql.fabric.server.py module or into the mysql.py.replication.py module. But
this would have resulted in tight coupling between the HA and the Server layers
and also would have resulted in a circular dependency, since the replication
module already has a reference to the server module.
"""
import logging

from mysql.fabric.server import Group,  MySQLServer
import mysql.fabric.replication as _replication
import mysql.fabric.errors as _errors

_LOGGER = logging.getLogger(__name__)

GROUP_REPLICATION_GROUP_NOT_FOUND_ERROR = "Group not found %s"
GROUP_REPLICATION_GROUP_MASTER_NOT_FOUND_ERROR = "Group master not found %s"
GROUP_MASTER_NOT_RUNNING = "Group master not running %s"

def start_group_slaves(master_group_id):
    """Start the slave groups for the given master group. The
    method will be used in the events that requires, a group, that
    has registered slaves to start them. An example would be
    enable shard, enable shard requires that a group start all
    the slaves that are registered with it.

    :param master_group_id: The master group ID. The ID belongs to the master
                            whose slaves need to be started.
    """
    # Fetch the master group corresponding to the master group
    # ID.
    master_group = Group.fetch(master_group_id)
    if master_group is None:
        raise _errors.GroupError(GROUP_REPLICATION_GROUP_NOT_FOUND_ERROR % \
                                                (master_group_id, ))

    # Setup replication with masters of the groups registered as master
    # groups. master_group.slave_group_ids contains the list of the group
    # IDs that are slaves to this master. Iterate through this list and start
    # replication with the registered slaves.
    for slave_group_id in master_group.slave_group_ids:
        slave_group = Group.fetch(slave_group_id)
        # Setup replication with the slave group.
        try:
            setup_group_replication(master_group_id, slave_group.group_id)
        except (_errors.GroupError, _errors.DatabaseError) as error:
            _LOGGER.warning(
                "Error while configuring group replication between "
                "(%s) and (%s): (%s).", master_group_id, slave_group.group_id,
                error
            )

def stop_group_slaves(master_group_id):
    """Stop the group slaves for the given master group. This will be used
    for use cases that required all the slaves replicating from this group to
    be stopped. An example use case would be disabling a shard.

    :param master_group_id: The master group ID.
    """
    master_group = Group.fetch(master_group_id)
    if master_group is None:
        raise _errors.GroupError \
        (GROUP_REPLICATION_GROUP_NOT_FOUND_ERROR % \
        (master_group_id, ))

    # Stop the replication on all of the registered slaves for the group.
    for slave_group_id in master_group.slave_group_ids:

        slave_group = Group.fetch(slave_group_id)
        # Fetch the Slave Group and the master of the Slave Group
        slave_group_master = MySQLServer.fetch(slave_group.master)
        if slave_group_master is None:
            _LOGGER.warning(
                GROUP_REPLICATION_GROUP_MASTER_NOT_FOUND_ERROR % \
                (slave_group.master, )
            )
            continue

        if not server_running(slave_group_master):
            # The server is already down. we cannot connect to it to stop
            # replication.
            continue

        try:
            slave_group_master.connect()
            _replication.stop_slave(slave_group_master, wait=True)
            # Reset the slave to remove the reference of the master so
            # that when the server is used as a slave next it does not
            # complaint about having a different master.
            _replication.reset_slave(slave_group_master, clean=True)
        except _errors.DatabaseError as error:
            # Server is not accessible, unable to connect to the server.
            _LOGGER.warning(
                "Error while unconfiguring group replication between "
                "(%s) and (%s): (%s).", master_group_id, slave_group.group_id,
                error
            )
            continue

def stop_group_slave(group_master_id,  group_slave_id,  clear_ref):
    """Stop the slave on the slave group. This utility method is the
    completement of the setup_group_replication method and is
    used to stop the replication on the slave group. Given a master group ID
    and the slave group ID the method stops the slave on the slave
    group and updates the references on both the master and the
    slave group.

    :param group_master_id: The id of the master group.
    :param group_slave_id: The id of the slave group.
    :param clear_ref: The parameter indicates if the stop_group_slave
                                needs to clear the references to the group's
                                slaves. For example when you do a disable
                                shard the shard group still retains the
                                references to its slaves, since when enabled
                                it needs to enable the replication.
    """
    master_group = Group.fetch(group_master_id)
    slave_group = Group.fetch(group_slave_id)

    if master_group is None:
        raise _errors.GroupError \
        (GROUP_REPLICATION_GROUP_NOT_FOUND_ERROR % (group_master_id, ))

    if slave_group is None:
        raise _errors.GroupError \
        (GROUP_REPLICATION_GROUP_NOT_FOUND_ERROR % (group_slave_id, ))

    slave_group_master = MySQLServer.fetch(slave_group.master)
    if slave_group_master is None:
        raise _errors.GroupError \
        (GROUP_REPLICATION_GROUP_MASTER_NOT_FOUND_ERROR %
          (slave_group.master, ))

    if not server_running(slave_group_master):
        #The server is already down. We cannot connect to it to stop
        #replication.
        return
    try:
        slave_group_master.connect()
    except _errors.DatabaseError:
        #Server is not accessible, unable to connect to the server.
        return

    #Stop replication on the master of the group and clear the references,
    #if clear_ref has been set.
    _replication.stop_slave(slave_group_master, wait=True)
    _replication.reset_slave(slave_group_master,  clean=True)
    if clear_ref:
        slave_group.remove_master_group_id()
        master_group.remove_slave_group_id(group_slave_id)

def setup_group_replication(group_master_id,  group_slave_id):
    """Sets up replication between the masters of the two groups and
    updates the references to the groups in each other.

    :param group_master_id: The group whose master will act as the master
                                             in the replication setup.
    :param group_slave_id: The group whose master will act as the slave in the
                                      replication setup.
    """
    group_master = Group.fetch(group_master_id)
    group_slave = Group.fetch(group_slave_id)

    if group_master is None:
        raise _errors.GroupError \
        (GROUP_REPLICATION_GROUP_NOT_FOUND_ERROR % (group_master_id, ))

    if group_slave is None:
        raise _errors.GroupError \
        (GROUP_REPLICATION_GROUP_NOT_FOUND_ERROR % (group_slave_id, ))

    if group_master.master is None:
        raise _errors.GroupError \
        (GROUP_REPLICATION_GROUP_MASTER_NOT_FOUND_ERROR % "")

    if group_slave.master is None:
        raise _errors.GroupError \
        (GROUP_REPLICATION_GROUP_MASTER_NOT_FOUND_ERROR % "")

    #Master is the master of the Global Group. We replicate from here to
    #the masters of all the shard Groups.
    master = MySQLServer.fetch(group_master.master)
    if master is None:
        raise _errors.GroupError \
        (GROUP_REPLICATION_GROUP_MASTER_NOT_FOUND_ERROR % \
        (group_master.master, ))

    #Get the master of the shard Group.
    slave = MySQLServer.fetch(group_slave.master)
    if slave is None:
        raise _errors.GroupError \
        (GROUP_REPLICATION_GROUP_MASTER_NOT_FOUND_ERROR % \
        (group_slave.master, ))

    if not server_running(master):
        #The server is already down. We cannot connect to it to setup
        #replication.
        raise _errors.GroupError \
        (GROUP_MASTER_NOT_RUNNING % (group_master.group_id, ))

    try:
        master.connect()
    except _errors.DatabaseError:
        #Server is not accessible, unable to connect to the server.
        raise _errors.GroupError \
            (GROUP_REPLICATION_GROUP_MASTER_NOT_FOUND_ERROR % \
            (group_slave.master, ))

    if not server_running(slave):
        #The server is already down. We cannot connect to it to setup
        #replication.
        raise _errors.GroupError \
            (GROUP_MASTER_NOT_RUNNING % (group_slave.group_id, ))

    try:
        slave.connect()
    except _errors.DatabaseError:
        #Server is not accessible, unable to connect to the server.
        raise _errors.GroupError \
            (GROUP_REPLICATION_GROUP_MASTER_NOT_FOUND_ERROR % \
            (group_master.master, ))

    _replication.stop_slave(slave, wait=True)

    #clear references to old masters in the slave
    _replication.reset_slave(slave,  clean=True)

    _replication.switch_master(slave, master, master.user, master.passwd)

    _replication.start_slave(slave, wait=True)

    try:
        group_master.add_slave_group_id(group_slave_id)
        group_slave.add_master_group_id(group_master_id)
    except _errors.DatabaseError:
        #If there is an error while adding a reference to
        #the slave group or a master group, it means that
        #the slave group was already added and the error
        #is happening because the group was already registered.
        #Ignore this error.
        pass

def server_running(server):
    """Check if the server is in the running state.

    :param server: The MySQLServer object whose mode needs to be checked.

    :return: True if server is in the running state.
             False otherwise.
    """
    if server.status in [MySQLServer.FAULTY]:
        return False
    return True



#
# Copyright (c) 2010, 2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""mysql.utilities"""

# Major, Minor, Patch, Status
VERSION = (1, 4, 3, 'GA', 0)
# Future versions will have to include only the X, Y (no Z).
WORKBENCH_VERSION = (6, 0, 0)

VERSION_STRING = "%s.%s.%s" % VERSION[0:3]
RELEASE_STRING = (
    VERSION_STRING +
    " (part of MySQL Workbench Distribution %s.%s.%s)" % WORKBENCH_VERSION)

COPYRIGHT = "2010, 2014 Oracle and/or its affiliates. All rights reserved."

COPYRIGHT_FULL = "Copyright (c) " + COPYRIGHT + """
This is a release of dual licensed MySQL Utilities. For the avoidance of
doubt, this particular copy of the software is released
under the version 2 of the GNU General Public License.
MySQL Utilities is brought to you by Oracle.
"""

LICENSE = "GPLv2"

VERSION_FRM = ("MySQL Utilities {program} version {RELEASE_STRING} \n"
               "License type: {LICENSE}".format(program="{program}",
                                                RELEASE_STRING=RELEASE_STRING,
                                                LICENSE=LICENSE))

LICENSE_FRM = (VERSION_FRM + "\n" + COPYRIGHT_FULL)
PYTHON_MIN_VERSION = (2, 6, 0)
PYTHON_MAX_VERSION = (3, 0, 0)
CONNECTOR_MIN_VERSION = (1, 0, 9)

# This dictionary has to be updated whenever a utility is added.
# the format to use is:
# '<utility_name>': (<PYTHON_MIN_VERSION>, <PYTHON_MAX_VERSION>)
AVAILABLE_UTILITIES = {
    'mysqlauditadmin': (),
    'mysqlauditgrep': (),
    'mysqldbcompare': (),
    'mysqldbcopy': (),
    'mysqldbexport': (),
    'mysqldbimport': (),
    'mysqldiff': (),
    'mysqldiskusage': (),
    'mysqlfailover': (),
    'mysqlfrm': (),
    'mysqlindexcheck': (),
    'mysqlmetagrep': (),
    'mysqlprocgrep': (),
    'mysqlreplicate': (),
    'mysqlrpladmin': (),
    'mysqlrplcheck': (),
    'mysqlrplms': (),
    'mysqlrplshow': (),
    'mysqlrplsync': (),
    'mysqlserverclone': (),
    'mysqlserverinfo': (),
    'mysqluc': (),
    'mysqluserclone': (),
}



#
# Copyright (c) 2011, 2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
This file contains the commands for checking consistency of two databases.
"""

from mysql.utilities.exception import UtilDBError, UtilError
from mysql.utilities.common.database import Database
from mysql.utilities.common.sql_transform import quote_with_backticks
from mysql.utilities.common.dbcompare import (diff_objects, get_common_objects,
                                              get_create_object,
                                              print_missing_list,
                                              server_connect,
                                              check_consistency,
                                              build_diff_list,
                                              DEFAULT_SPAN_KEY_SIZE)
from mysql.utilities.common.server import connect_servers

_PRINT_WIDTH = 75
_ROW_FORMAT = "# {0:{1}} {2:{3}} {4:{5}} {6:{7}} {8:{9}}"
_RPT_FORMAT = "{0:{1}} {2:{3}}"

_ERROR_DB_DIFF = "The object definitions do not match."
_ERROR_DB_MISSING = "The database {0} does not exist."
_ERROR_OBJECT_LIST = "The list of objects differs among database {0} and {1}."
_ERROR_ROW_COUNT = "Row counts are not the same among {0} and {1}.\n#"
_ERROR_DB_MISSING_ON_SERVER = "The database {0} on {1} does not exist on {2}."

_DEFAULT_OPTIONS = {
    "quiet": False,
    "verbosity": 0,
    "difftype": "differ",
    "run_all_tests": False,
    "width": 75,
    "no_object_check": False,
    "no_diff": False,
    "no_row_count": False,
    "no_data": False,
    "transform": False,
    "span_key_size": DEFAULT_SPAN_KEY_SIZE
}


class _CompareDBReport(object):
    """Print compare database report
    """

    def __init__(self, options):
        """Constructor

        options[in]    options for class
            width[in]      Width of report
            quiet[in]      If true, do not print commentary
                           (default = False)
        """
        self.width = options.get('width', _PRINT_WIDTH) - 2  # for '# '
        self.quiet = options.get('quiet', False)
        self.type_width = 9
        self.oper_width = 7
        self.desc_width = self.width - self.type_width - \
            (3 * self.oper_width) - 4

    def print_heading(self):
        """Print heading for database consistency
        """
        # Skip if quiet
        if self.quiet:
            return
        # Set the variable width global parameters here
        print _ROW_FORMAT.format(' ', self.type_width,
                                 ' ', self.desc_width,
                                 "Defn", self.oper_width,
                                 "Row", self.oper_width,
                                 "Data", self.oper_width)
        print _ROW_FORMAT.format("Type", self.type_width,
                                 "Object Name", self.desc_width,
                                 "Diff", self.oper_width,
                                 "Count", self.oper_width,
                                 "Check", self.oper_width)
        print "# %s" % ('-' * self.width),

    def report_object(self, obj_type, description):
        """Print the object type and description field

        obj_type[in]      type of the object(s) described
        description[in]   description of object(s)
        """
        # Skip if quiet
        if self.quiet:
            return
        print "\n#", _RPT_FORMAT.format(obj_type, self.type_width,
                                        description, self.desc_width),

    def report_state(self, state):
        """Print the results of a test.

        state[in]         state of the test
        """
        # Skip if quiet
        if self.quiet:
            return
        print "{0:<{1}}".format(state, self.oper_width),

    @staticmethod
    def report_errors(errors):
        """Print any errors encountered.

        errors[in]        list of strings to print
        """
        if len(errors) > 0:
            print "\n#"
        for line in errors:
            print line


def _check_databases(server1, server2, db1, db2, options):
    """Check databases

    server1[in]       first server Server instance
    server2[in]       second server Server instance
    db1[in]           first database
    db2[in]           second database
    options[in]       options dictionary

    Returns tuple - Database class instances for databases
    """

    # Check database create for differences
    if not options['no_diff']:
        # temporarily make the diff quiet to retrieve errors
        new_opt = {}
        new_opt.update(options)
        new_opt['quiet'] = True          # do not print messages
        new_opt['suppress_sql'] = True   # do not print SQL statements either
        res = diff_objects(server1, server2, db1, db2, new_opt, 'DATABASE')
        if res is not None:
            for row in res:
                print row
            print
            if not options['run_all_tests']:
                raise UtilError(_ERROR_DB_DIFF)


def _check_objects(server1, server2, db1, db2,
                   db1_conn, db2_conn, options):
    """Check number of objects

    server1[in]       first server Server instance
    server2[in]       second server Server instance
    db1[in]           first database
    db2[in]           second database
    db1_conn[in]      first Database instance
    db2_conn[in]      second Database instance
    options[in]       options dictionary

    Returns list of objects in both databases
    """

    differs = False

    # Check for same number of objects
    in_both, in_db1, in_db2 = get_common_objects(server1, server2,
                                                 db1, db2, False, options)
    in_both.sort()
    if not options['no_object_check']:
        server1_str = "server1." + db1
        if server1 == server2:
            server2_str = "server1." + db2
        else:
            server2_str = "server2." + db2
        if len(in_db1) or len(in_db2):
            if options['run_all_tests']:
                if len(in_db1) > 0:
                    differs = True
                    print_missing_list(in_db1, server1_str, server2_str)
                    print "#"
                if len(in_db2) > 0:
                    differs = True
                    print_missing_list(in_db2, server2_str, server1_str)
                    print "#"
            else:
                differs = True
                raise UtilError(_ERROR_OBJECT_LIST.format(db1, db2))

    # If in verbose mode, show count of object types.
    if options['verbosity'] > 1:
        objects = {
            'TABLE': 0,
            'VIEW': 0,
            'TRIGGER': 0,
            'PROCEDURE': 0,
            'FUNCTION': 0,
            'EVENT': 0,
        }
        for item in in_both:
            obj_type = item[0]
            objects[obj_type] += 1
        print "Looking for object types table, view, trigger, procedure," + \
              " function, and event."
        print "Object types found common to both databases:"
        for obj in objects:
            print " {0:>12} : {1}".format(obj, objects[obj])

    return (in_both, differs)


def _compare_objects(server1, server2, obj1, obj2, reporter, options,
                     object_type):
    """Compare object definitions and produce difference

    server1[in]       first server Server instance
    server2[in]       second server Server instance
    obj1[in]          first object
    obj2[in]          second object
    reporter[in]      database compare reporter class instance
    options[in]       options dictionary
    object_type[in]   type of the objects to be compared (e.g., TABLE,
                      PROCEDURE, etc.).

    Returns list of errors
    """

    errors = []
    if not options['no_diff']:
        # For each database, compare objects
        # temporarily make the diff quiet to retrieve errors
        new_opt = {}
        new_opt.update(options)
        new_opt['quiet'] = True          # do not print messages
        new_opt['suppress_sql'] = True   # do not print SQL statements either
        res = diff_objects(server1, server2, obj1, obj2, new_opt, object_type)
        if res is not None:
            reporter.report_state('FAIL')
            errors.extend(res)
            if not options['run_all_tests']:
                raise UtilError(_ERROR_DB_DIFF)
        else:
            reporter.report_state('pass')
    else:
        reporter.report_state('SKIP')

    return errors


def _check_row_counts(server1, server2, obj1, obj2, reporter, options):
    """Compare row counts for tables

    server1[in]       first server Server instance
    server2[in]       second server Server instance
    obj1[in]          first object
    obj2[in]          second object
    reporter[in]      database compare reporter class instance
    options[in]       options dictionary

    Returns list of errors
    """
    errors = []
    if not options['no_row_count']:
        rows1 = server1.exec_query("SELECT COUNT(*) FROM " + obj1)
        rows2 = server2.exec_query("SELECT COUNT(*) FROM " + obj2)
        if rows1 != rows2:
            reporter.report_state('FAIL')
            msg = _ERROR_ROW_COUNT.format(obj1, obj2)
            if not options['run_all_tests']:
                raise UtilError(msg)
            else:
                errors.append("# %s" % msg)
        else:
            reporter.report_state('pass')
    else:
        reporter.report_state('SKIP')

    return errors


def _check_data_consistency(server1, server2, obj1, obj2, reporter, options):
    """Check data consistency

    server1[in]       first server Server instance
    server2[in]       second server Server instance
    obj1[in]          first object
    obj2[in]          second object
    reporter[in]      database compare reporter class instance
    options[in]       options dictionary

    Returns list of errors debug_msgs
    """
    direction = options.get('changes-for', 'server1')
    reverse = options.get('reverse', False)

    errors = []
    debug_msgs = []
    diff_server1 = []
    diff_server2 = []
    # For each table, do row data consistency check
    if not options['no_data']:
        reporter.report_state('-')
        try:
            # Do the comparison based on direction
            if direction == 'server1' or reverse:
                diff_server1 = check_consistency(server1, server2,
                                                 obj1, obj2, options,
                                                 diag_msgs=debug_msgs,
                                                 reporter=reporter)
            if direction == 'server2' or reverse:
                diff_server2 = check_consistency(server2, server1,
                                                 obj2, obj1, options,
                                                 diag_msgs=debug_msgs,
                                                 reporter=reporter)

            # if no differences, return
            if (diff_server1 is None and diff_server2 is None) or \
                    (not reverse and direction == 'server1' and
                     diff_server1 is None) or \
                    (not reverse and direction == 'server2' and
                     diff_server2 is None):
                return errors, debug_msgs

            # Build diff list
            new_opts = options.copy()
            new_opts['data_diff'] = True
            if direction == 'server1':
                diff_list = build_diff_list(diff_server1, diff_server2,
                                            diff_server1, diff_server2,
                                            'server1', 'server2', new_opts)
            else:
                diff_list = build_diff_list(diff_server2, diff_server1,
                                            diff_server2, diff_server1,
                                            'server2', 'server1', new_opts)
            if diff_list:
                errors = diff_list
        except UtilError, e:
            if e.errmsg.endswith("not have an usable Index or primary key."):
                reporter.report_state('SKIP')
                errors.append("# {0}".format(e.errmsg))
            else:
                reporter.report_state('FAIL')
                if not options['run_all_tests']:
                    print
                    raise e
                else:
                    errors.append(e.errmsg)

    else:
        reporter.report_state('SKIP')

    return errors, debug_msgs


def _check_option_defaults(options):
    """Set the defaults for options if they are not set.

    This prevents users from calling the method and its subordinates
    with missing options.
    """

    for opt_name in _DEFAULT_OPTIONS:
        if not opt_name in options:
            options[opt_name] = _DEFAULT_OPTIONS[opt_name]


def database_compare(server1_val, server2_val, db1, db2, options):
    """Perform a consistency check among two databases

    This method performs a database consistency check among two databases which
    ensures the databases exist, the objects match in number and type, the row
    counts match for all tables, and the data for each matching tables is
    consistent.

    If any errors or differences are found, the operation stops and the
    difference is printed.

    The following steps are therefore performed:

    1) check to make sure the databases exist and are the same definition
    2) check to make sure the same objects exist in each database
    3) for each object, ensure the object definitions match among the databases
    4) for each table, ensure the row counts are the same
    5) for each table, ensure the data is the same

    By default, the operation stops on any failure of any test. The caller can
    override this behavior by specifying run_all_tests = True in the options
    dictionary.

    TODO:   allow the user to skip object types (e.g. --skip-triggers, et. al.)

    server1_val[in]    a dictionary containing connection information for the
                       first server including:
                       (user, password, host, port, socket)
    server2_val[in]    a dictionary containing connection information for the
                       second server including:
                       (user, password, host, port, socket)
    db1[in]            the first database in the compare
    db2[in]            the second database in the compare
    options[in]        a dictionary containing the options for the operation:
                       (quiet, verbosity, difftype, run_all_tests)

    Returns bool True if all object match, False if partial match
    """

    _check_option_defaults(options)

    # Connect to servers
    server1, server2 = server_connect(server1_val, server2_val,
                                      db1, db2, options)

    # Check to see if databases exist
    db1_conn = Database(server1, db1, options)
    if not db1_conn.exists():
        raise UtilDBError(_ERROR_DB_MISSING.format(db1))

    db2_conn = Database(server2, db2, options)
    if not db2_conn.exists():
        raise UtilDBError(_ERROR_DB_MISSING.format(db2))

    # Print a different message is server2 is not defined
    if not server2_val:
        message = "# Checking databases {0} and {1} on server1\n#"
    else:
        message = "# Checking databases {0} on server1 and {1} on server2\n#"
    print(message.format(db1_conn.db_name, db2_conn.db_name))

    # Check for database existence and CREATE differences
    _check_databases(server1, server2, db1_conn.q_db_name, db2_conn.q_db_name,
                     options)

    # Get common objects and report discrepancies
    (in_both, differs) = _check_objects(server1, server2, db1, db2,
                                        db1_conn, db2_conn, options)
    success = not differs

    reporter = _CompareDBReport(options)
    reporter.print_heading()

    # Remaining operations can occur in a loop one for each object.
    for item in in_both:
        error_list = []
        debug_msgs = []
        # Set the object type
        obj_type = item[0]

        q_obj1 = "{0}.{1}".format(quote_with_backticks(db1),
                                  quote_with_backticks(item[1][0]))
        q_obj2 = "{0}.{1}".format(quote_with_backticks(db2),
                                  quote_with_backticks(item[1][0]))

        reporter.report_object(obj_type, item[1][0])

        # Check for differences in CREATE
        errors = _compare_objects(server1, server2, q_obj1, q_obj2,
                                  reporter, options, obj_type)
        error_list.extend(errors)

        # Check row counts
        if obj_type == 'TABLE':
            errors = _check_row_counts(server1, server2, q_obj1, q_obj2,
                                       reporter, options)
            if len(errors) != 0:
                error_list.extend(errors)
        else:
            reporter.report_state("-")

        # Check data consistency for tables
        if obj_type == 'TABLE':
            errors, debug_msgs = _check_data_consistency(server1, server2,
                                                         q_obj1, q_obj2,
                                                         reporter, options)
            if len(errors) != 0:
                error_list.extend(errors)
        else:
            reporter.report_state("-")

        if options['verbosity'] > 0:
            print
            get_create_object(server1, q_obj1, options, obj_type)
            get_create_object(server2, q_obj2, options, obj_type)

        if debug_msgs and options['verbosity'] > 2:
            reporter.report_errors(debug_msgs)

        reporter.report_errors(error_list)

        # Fail if errors are found
        if error_list:
            success = False

    return success


def compare_all_databases(server1_val, server2_val, exclude_list, options):
    """Perform a consistency check among all common databases on the servers

    This method gets all databases from the servers, prints any missing
    databases and performs a consistency check among all common databases.

    If any errors or differences are found, the operation will print the
    difference and continue.

    This method will return None if no databases to compare.
    """

    success = True

    # Connect to servers
    conn_options = {
        "quiet": options.get("quiet", False),
        "src_name": "server1",
        "dest_name": "server2",

    }
    server1, server2 = connect_servers(server1_val, server2_val, conn_options)

    # Get all databases, except those used in --exclude
    get_dbs_query = """
        SELECT SCHEMA_NAME
        FROM INFORMATION_SCHEMA.SCHEMATA
        WHERE SCHEMA_NAME != 'INFORMATION_SCHEMA'
        AND SCHEMA_NAME != 'PERFORMANCE_SCHEMA'
        AND SCHEMA_NAME != 'mysql'
        {0}"""
    conditions = ""
    if exclude_list:
        # Add extra where to exclude databases in exclude_list
        operator = 'REGEXP' if options['use_regexp'] else 'LIKE'
        conditions = "AND {0}".format(" AND ".join(
            ["SCHEMA_NAME NOT {0} '{1}'".format(operator, db)
             for db in exclude_list])
        )

    server1_dbs = set(
        [db[0] for db in server1.exec_query(get_dbs_query.format(conditions))]
    )
    server2_dbs = set(
        [db[0] for db in server2.exec_query(get_dbs_query.format(conditions))]
    )

    # Check missing databases
    if options['changes-for'] == 'server1':
        diff_dbs = server1_dbs.difference(server2_dbs)
        for db in diff_dbs:
            msg = _ERROR_DB_MISSING_ON_SERVER.format(db, "server1", "server2")
            print("# {0}".format(msg))
    else:
        diff_dbs = server2_dbs.difference(server1_dbs)
        for db in diff_dbs:
            msg = _ERROR_DB_MISSING_ON_SERVER.format(db, "server2", "server1")
            print("# {0}".format(msg))

    # Compare databases in common
    common_dbs = server1_dbs.intersection(server2_dbs)
    if common_dbs:
        print("# Comparing databases: {0}".format(", ".join(common_dbs)))
    else:
        success = None
    for db in common_dbs:
        try:
            res = database_compare(server1_val, server2_val, db, db, options)
            if not res:
                success = False
            print("\n")
        except UtilError as err:
            print("ERROR: {0}\n".format(err.errmsg))
            success = False

    return success



#
# Copyright (c) 2013,2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""Define checkpoint routines that are responsible for providing the necessary
means so that one can guarantee consistency in the case of a failure.
"""

import uuid as _uuid
import time
import pickle as pickle
import sys
import logging

import mysql.fabric.persistence as _persistence
import mysql.fabric.errors as _errors


class Checkpoint(_persistence.Persistable):
    """This is responsible for keeping track of a procedure execution
    providing the necessary means to register the jobs that are triggered
    on its behalf.

    While running the recovery procedure, the entry with the greatest
    sequence is retrieved and the appropriate actions executed to put the
    system in a consistent state.

    See in what follows, a figure that briefly depicts what happens:

    .. seqdiag::

      diagram {
      activation = none;
      === Register "action" and Create "procedure" ===
      trigger -> executor;
      executor -> procedure [ label = "register(action)" ];
      executor <- procedure;
      trigger <- executor;
      === Execute "action" ===
      executor -> checkpoint [ label = "begin(action)" ];
      executor <- checkpoint;
      executor -> action [ label = "execute(action)" ];
      executor <- action;
      executor -> checkpoint [ label = "finish(action)" ];
      executor <- checkpoint;
      }

    :param proc_uuid: Procedure uuid.
    :param lockable_objects: Set of objects to be locked by the concurrency
                             control mechanism.
    :param job_uuid: Job uuid.
    :param action_fqn: Reference to the main function.
    :type action_fqn: String
    :param param_args: List with non-keyworded arguments to the
                       function(s).
    :param param_kwargs: Dictionary with neyworded arguments to the
                         function(s).
    :param started: Timestamp that identifies when the function
                    has started.
    :param finished: Timestamp that identifies when the function
                     has completed.
    :param sequence: Identify the order of the job within a procedure.
    """
    # SQL Statement for creating the table checkpoint which is used to
    # keep track of execution (i.e. jobs).
    CREATE_CHECKPOINTS = (
        "CREATE TABLE checkpoints (proc_uuid VARCHAR(40) NOT NULL, "
        "lockable_objects BLOB NULL, job_uuid VARCHAR(40) NOT NULL, "
        "sequence INTEGER NOT NULL, action_fqn TEXT NOT NULL, "
        "param_args BLOB NULL, param_kwargs BLOB NULL, "
        "started DOUBLE(16, 6) NULL, finished DOUBLE(16, 6) NULL, "
        "CONSTRAINT pk_checkpoint PRIMARY KEY (proc_uuid, job_uuid))"
        )

    #SQL statement for inserting a new checkpoint into the table.
    INSERT_CHECKPOINT = (
        "INSERT INTO checkpoints(proc_uuid, lockable_objects, job_uuid, "
        "sequence, action_fqn, param_args, param_kwargs) "
        "VALUES (%s, %s, %s, %s, %s, %s, %s)"
        )

    #SQL statement for updating the started time.
    UPDATE_START_CHECKPOINT = ("UPDATE checkpoints set started = %s WHERE "
        "proc_uuid = %s and job_uuid = %s"
        )

    #SQL statement for updating the finished time.
    UPDATE_FINISH_CHECKPOINT = ("UPDATE checkpoints set finished = %s WHERE "
        "proc_uuid = %s and job_uuid = %s"
        )

    #SQL statement for deleting the checkpoint executed on behalf of
    #a procedure.
    DELETE_CHECKPOINTS = ("DELETE FROM checkpoints WHERE proc_uuid = %s")

    #SQL statement for retrieving the checkpoints stored on behalf of a
    #procedure.
    QUERY_CHECKPOINTS = ("SELECT proc_uuid, lockable_objects, job_uuid, "
        "sequence, action_fqn, param_args, param_kwargs, started, finished "
        "FROM checkpoints WHERE proc_uuid = %s"
        )

    #SQL statement for retrieving all occurrences in the checkptoint which
    #is used for recovery.
    QUERY_UNFINISHED_CHECKPOINTS = (
        "SELECT chk_info.proc_uuid, chk_info.lockable_objects, "
        "chk_info.job_uuid, chk_info.sequence, chk_info.action_fqn, "
        "chk_info.param_args, chk_info.param_kwargs, "
        "chk_info.started, chk_info.finished FROM "
        "(SELECT proc_uuid, max(sequence) as sequence FROM checkpoints "
        "WHERE started is NOT NULL AND finished is NULL GROUP BY proc_uuid) "
        "AS chk_core INNER JOIN "
        "(SELECT proc_uuid, lockable_objects, job_uuid, sequence, action_fqn, "
        "param_args, param_kwargs, started, finished FROM checkpoints) "
        "AS chk_info ON chk_info.proc_uuid = chk_core.proc_uuid and "
        "chk_info.sequence = chk_core.sequence"
        )

    QUERY_REGISTERED_CHECKPOINTS = (
        "SELECT proc_uuid, lockable_objects, job_uuid, sequence, "
        "action_fqn, param_args, param_kwargs, started, finished "
        "FROM checkpoints WHERE finished "
        "is NULL ORDER BY proc_uuid, sequence"
        )

    QUERY_FINISHED_CHECKPOINTS = (
        "SELECT DISTINCT proc_uuid FROM checkpoints WHERE proc_uuid IN "
        "(SELECT DISTINCT chk_info.proc_uuid FROM checkpoints as chk_info "
        "WHERE chk_info.finished is NOT NULL) and proc_uuid NOT IN "
        "(SELECT DISTINCT chk_info.proc_uuid FROM checkpoints as chk_info "
        "WHERE chk_info.finished is NULL)"
        )

    def __init__(self, proc_uuid, lockable_objects, job_uuid, sequence,
                 action_fqn, param_args, param_kwargs, started=None,
                 finished=None):
        """Constructor for Checkpoint object.
        """
        super(Checkpoint, self).__init__()
        assert(isinstance(proc_uuid, _uuid.UUID))
        assert(isinstance(job_uuid, _uuid.UUID))
        assert(isinstance(action_fqn, basestring))
        assert(started is None or isinstance(started, float))
        assert(finished is None or isinstance(started, float))
        self.__proc_uuid = proc_uuid
        self.__job_uuid = job_uuid
        self.__action_fqn = action_fqn
        self.__param_args = param_args
        self.__param_kwargs = param_kwargs
        self.__do_action = Checkpoint.get_do_action(action_fqn)
        self.__undo_action = Checkpoint.get_undo_action(action_fqn)
        self.__sequence = sequence
        self.__started = started
        self.__finished = finished
        self.__lockable_objects = lockable_objects

    @property
    def lockable_objects(self):
        """Return the set of lockable objects.
        """
        return self.__lockable_objects

    @property
    def proc_uuid(self):
        """Return proc_uuid.
        """
        return self.__proc_uuid

    @property
    def job_uuid(self):
        """Return job_uuid.
        """
        return self.__job_uuid

    @property
    def param_args(self):
        """Return param_args.
        """
        return self.__param_args

    @property
    def param_kwargs(self):
        """Return param_kwargs.
        """
        return self.__param_kwargs

    @property
    def do_action(self):
        """Return do_action.
        """
        return self.__do_action

    @property
    def undo_action(self):
        """Return undo_action.
        """
        return self.__undo_action

    @property
    def started(self):
        """Return started.
        """
        return self.__started

    @property
    def finished(self):
        """Return finished.
        """
        return self.__finished

    @property
    def sequence(self):
        """Return sequence.
        """
        return self.__sequence

    def register(self, persister=None):
        """Register that an action has been registered.
        """
        param_args, param_kwargs, lockable_objects = \
            Checkpoint.serialize(self.__param_args, self.__param_kwargs,
                                 self.__lockable_objects)
        persister.exec_stmt(Checkpoint.INSERT_CHECKPOINT,
            {"params":(str(self.__proc_uuid), lockable_objects,
            str(self.__job_uuid), self.__sequence, self.__action_fqn,
            param_args, param_kwargs)}
            )

    def begin(self, persister=None):
        """Register that an action is about to start.
        """
        started = time.time()
        persister.exec_stmt(Checkpoint.UPDATE_START_CHECKPOINT,
            {"params":(started, str(self.__proc_uuid),
            str(self.__job_uuid))}
            )
        self.__started = started

    def finish(self, persister=None):
        """Register that a job has finished.

        :param job_uuid: Job uuid.
        :param persister: The DB server that can be used to access the
                          state store.
        """
        finished = time.time()
        persister.exec_stmt(Checkpoint.UPDATE_FINISH_CHECKPOINT,
            {"params":(finished, str(self.__proc_uuid),
            str(self.__job_uuid))}
            )
        self.__finished = finished

    @staticmethod
    def _create_object_from_row(row):
        """Create a Checkpoint object from a retrieved row.

        :param row: Checkpoint row.
        :type row: Tuple.
        :return: Return a Checkpoint object.
        """
        (proc_uuid, lockable_objects, job_uuid, sequence, action_fqn,
         param_args, param_kwargs, started, finished) = row
        param_args, param_kwargs, lockable_objects = \
             Checkpoint.deserialize(param_args, param_kwargs, lockable_objects)
        checkpoint = Checkpoint(
            _uuid.UUID(proc_uuid), lockable_objects, _uuid.UUID(job_uuid),
            sequence, action_fqn, param_args, param_kwargs,
            started, finished
            )
        return checkpoint

    @staticmethod
    def unfinished(persister=None):
        """Return unfinished procedures.

        :param persister: The DB server that can be used to access the
                          state store.
        :return: Set of procedures that haven't finished.
        :rtype: set(Checkpoint, ...)
        """
        checkpoints = set()
        rows = persister.exec_stmt(Checkpoint.QUERY_UNFINISHED_CHECKPOINTS,
                                   {"raw": False})
        for row in rows:
            checkpoints.add(Checkpoint._create_object_from_row(row))
        return checkpoints

    @staticmethod
    def registered(persister=None):
        """Return registered procedures.

        :param persister: The DB server that can be used to access the
                          state store.
        :return: Set of procedures that were registered.
        :rtype: set(Checkpoint, ...)
        """
        checkpoints = set()
        rows = persister.exec_stmt(Checkpoint.QUERY_REGISTERED_CHECKPOINTS,
                                   {"raw": False})
        for row in rows:
            checkpoints.add(Checkpoint._create_object_from_row(row))
        return checkpoints

    @staticmethod
    def fetch(proc_uuid, persister=None):
        """Return the object corresponding to the proc_uuid.

        :param proc_uuid: Procedure uuid.
        :param persister: The DB server that can be used to access the
                          state store.
        :return: Checkpoint object that corresponds to the proc_uuid or
                 None if one is not found.
        """
        checkpoints = set()
        assert(isinstance(proc_uuid, _uuid.UUID))

        rows = persister.exec_stmt(Checkpoint.QUERY_CHECKPOINTS,
            {"raw": False, "params":(str(proc_uuid), )}
            )

        if rows:
            for row in rows:
                checkpoint = Checkpoint._create_object_from_row(row)
                checkpoints.add(checkpoint)
        return checkpoints

    @staticmethod
    def remove(checkpoint, persister=None):
        """Remove the object from the persistent store.

        :param checkpoint: Checkpoint object.
        :param persister: The DB server that can be used to access the
                          state store.
        """
        assert(isinstance(checkpoint, Checkpoint))
        persister.exec_stmt(Checkpoint.DELETE_CHECKPOINTS,
            {"params":(str(checkpoint.proc_uuid), )}
            )

    @staticmethod
    def cleanup(persister=None):
        """Remove all the checkpoints which are related to procedures
        that have finished the execution but did not get the chance
        of calling remove.

        :param persister: The DB server that can be used to access the
                          state store.
        """
        rows = persister.exec_stmt(Checkpoint.QUERY_FINISHED_CHECKPOINTS,
            {"raw": False}
            )
        if rows:
            for row in rows:
                persister.exec_stmt(Checkpoint.DELETE_CHECKPOINTS,
                    {"params":(row[0], )}
                    )

    @staticmethod
    def create(persister=None):
        """Create the objects(tables) that represent Checkpoint information
        in the state store.

        :param persister: The DB server that can be used to access the
                          state store.
        """
        persister.exec_stmt(Checkpoint.CREATE_CHECKPOINTS)

    @staticmethod
    def get_do_action(action_fqn):
        """Get a reference to a main action.

        :param action_fqn: Fully qualified function name, i.e. module.name.
        :return: Reference to an action if there is any.
        :rtype: Callable or None.
        """
        module , name = action_fqn.rsplit(".", 1)
        try:
            return getattr(sys.modules[module], name)
        except (AttributeError, KeyError, NameError):
            return None

    @staticmethod
    def get_undo_action(action_fqn):
        """Get a reference to a main action.

        :param action_fqn: Fully qualified function name, i.e. module.name.
        :return: Reference to an action if there is any.
        :rtype: Callable or None.
        """
        module , name = action_fqn.rsplit(".", 1)
        try:
            return getattr(sys.modules[module], name).undo_function
        except (AttributeError, KeyError, NameError):
            return None

    @staticmethod
    def serialize(param_args, param_kwargs, lockable_objects):
        """Serialize the non-keyworded and keyworded parameters using Pickle.
        It is worth noticing that it does not check the type of the objects
        that are being serialize and it is up to the user to do so.

        :param param_args: List with non-keyworded arguments to the
                           function(s).
        :param param_kwargs: Dictionary with neyworded arguments to the
                             function(s).
        :param lockable_objects: Set of objects to be locked by the concurrency
                                 control mechanism.
        :return: Return a tuple with the parameters serialized.
        :rtype: (serialized_args, serialized_kwargs,
                serialized_lockable_objects).
        """
        s_param_args = pickle.dumps(param_args)
        s_param_kwargs = pickle.dumps(param_kwargs)
        s_lockable_objects = pickle.dumps(lockable_objects)
        return s_param_args, s_param_kwargs, s_lockable_objects

    @staticmethod
    def deserialize(param_args, param_kwargs, lockable_objects):
        """Deserialize the non-keyworded and keyworded parameters using Pickle.

        :param param_args: Serialized list with non-keyworded arguments to the
                           function(s).
        :param param_kwargs: Serialized dictionary with neyworded arguments to
                             the function(s).
        :param lockable_objects: Set of objects to be locked by the concurrency
                                 control mechanism.
        :return: Return a tuple with the parameters deserialized.
        :rtype: (args, kwargs, lockable_objects).
        """
        ds_param_args = pickle.loads(param_args)
        ds_param_kwargs = pickle.loads(param_kwargs)
        ds_lockable_objects = pickle.loads(lockable_objects)
        return ds_param_args, ds_param_kwargs, ds_lockable_objects

    @staticmethod
    def is_recoverable(action):
        """Check if an action is recoverable or not. This means that it is
        possible to get a reference to the action given its module name and
        and its name.

        :param action: Callable.
        :return: True if it is recoverable. Otherwise, return False.
        """
        try:
            action_fqn = action.__module__ + "." + action.__name__
            return Checkpoint.get_do_action(action_fqn) is not None
        except AttributeError:
            return False

    def __eq__(self,  other):
        """Two entries are equal if they have the same proc_uuid and job_uuid.
        """
        return isinstance(other, Checkpoint) and \
               self.__proc_uuid == other.proc_uuid and \
               self.__job_uuid == other.job_uuid

    def __hash__(self):
        """A Checkpoint is hashable through its proc_uuid and job_uuid.
        """
        return hash(self.__proc_uuid) ^ hash(self.__job_uuid)

_LOGGER = logging.getLogger(__name__)

def register(jobs, transaction):
    """Atomically register jobs.

    :param jobs: List of jobs to be registered.
    :param transaction: Whether there is transaction context or not.
    """
    assert(isinstance(jobs, list))

    persister = _persistence.current_persister()
    if not transaction:
        persister.begin()

    try:
        for job in jobs:
            if job.is_recoverable:
                job.checkpoint.register()

    except _errors.DatabaseError:
        try:
            if not transaction:
                persister.rollback()
        except _errors.DatabaseError as error:
            _LOGGER.error(
                "Error rolling back registered jobs.", exc_info=error
            )
        raise

    else:
        try:
            # Currently, if the commit fails, we are not sure whether the
            # changes have succeeded or not. This is something that needs
            # to be improved in the near future.
            if not transaction:
                persister.commit()
        except _errors.DatabaseError as error:
            _LOGGER.error(
                "Error committing registered jobs.", exc_info=error
            )
            raise



#!/usr/bin/env python
#
# Copyright (c) 2012, 2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
This file contains the replication slave administration utility. It is used to
perform replication operations on one or more slaves.
"""

from mysql.utilities.common.tools import check_python_version

# Check Python version compatibility
check_python_version()

import logging
import os.path
import signal
import sys

from mysql.utilities import VERSION_FRM
from mysql.utilities.exception import UtilError, UtilRplError
from mysql.utilities.command.rpl_admin import RplCommands, purge_log
from mysql.utilities.common.messages import SCRIPT_THRESHOLD_WARNING
from mysql.utilities.common.server import check_hostname_alias
from mysql.utilities.common.tools import check_connector_python
from mysql.utilities.common.topology import parse_topology_connections
from mysql.utilities.common.options import (add_verbosity, add_rpl_user,
                                            add_failover_options,
                                            check_server_lists,
                                            license_callback,
                                            UtilitiesParser)


# Constants
NAME = "MySQL Utilities - mysqlfailover "
DESCRIPTION = ("mysqlfailover - automatic replication health monitoring and "
               "failover")
USAGE = "%prog --master=root@localhost --discover-slaves-login=root " + \
        "--candidates=root@host123:3306,root@host456:3306 "
_DATE_FORMAT = '%Y-%m-%d %H:%M:%S %p'
_DATE_LEN = 22

# Check for connector/python
if not check_connector_python():
    sys.exit(1)

if __name__ == '__main__':
    # Setup a terminal signal handler for SIGNIT
    # Must use SetConsoleCtrlHandler function on Windows!
    # If posix, save old terminal settings so we can restore them on exit.
    try:
        # Only valid for *nix systems.
        import termios
        old_terminal_settings = termios.tcgetattr(sys.stdin)
    except:
        # Ok to fail for Windows
        pass

    def set_signal_handler(func):
        """Set the signal handler.
        """
        # If posix, restore old terminal settings.
        if os.name == "nt":
            from ctypes import windll
            windll.kernel32.SetConsoleCtrlHandler(func, True)
        # Install SIGTERM signal handler
        else:
            signal.signal(signal.SIGTERM, func)

    # If ctypes present, we have Windows so define the exit with decorators
    try:
        import ctypes

        @ctypes.WINFUNCTYPE(ctypes.c_int, ctypes.c_uint)
        def on_exit(signal, func=None):
            """Override the on_exit callback.
            """
            logging.info("Failover console stopped with SIGTERM.")
            sys.exit(0)
    except:
        def on_exit(signal, func=None):
            """Override the on_exit callback.
            """
            if os.name == "posix":
                termios.tcsetattr(sys.stdin, termios.TCSADRAIN,
                                  old_terminal_settings)
            logging.info("Failover console stopped with SIGTERM.")
            sys.exit(0)

    set_signal_handler(on_exit)

    # Setup the command parser
    program = os.path.basename(sys.argv[0]).replace(".py", "")
    parser = UtilitiesParser(
        version=VERSION_FRM.format(program=program),
        description=DESCRIPTION,
        usage=USAGE,
        add_help_option=False,
        prog=program
    )

    # Default option to provide help information
    parser.add_option("--help", action="help",
                      help="display this help message and exit")

    # Add --License option
    parser.add_option("--license", action='callback',
                      callback=license_callback,
                      help="display program's license and exit")

    # Setup utility-specific options:
    add_failover_options(parser)

    # Interval for continuous mode
    parser.add_option("--interval", "-i", action="store", dest="interval",
                      type="int", default="15",
                      help="interval in seconds for polling the master for "
                           "failure and reporting health. Default = 15 "
                           "seconds. Lowest value is 5 seconds.")

    # Add failover modes
    parser.add_option("--failover-mode", "-f", action="store",
                      dest="failover_mode", type="choice", default="auto",
                      choices=["auto", "elect", "fail"],
                      help="action to take when the master fails. 'auto' = "
                           "automatically fail to best slave, 'elect' = fail "
                           "to candidate list or if no candidate meets "
                           "criteria fail, 'fail' = take no action and stop "
                           "when master fails. Default = 'auto'.")

    # Add failover detection extension point
    parser.add_option("--exec-fail-check", action="store", dest="exec_fail",
                      type="string", default=None,
                      help="name of script to execute on each interval "
                           "to invoke failover")

    # Add force to override registry entry
    parser.add_option("--force", action="store_true", dest="force",
                      default=False,
                      help="override the registration check on master for "
                           "multiple instances of the console monitoring the "
                           "same master.")

    # Add refresh script external point
    parser.add_option("--exec-post-failover", action="store",
                      dest="exec_post_fail", type="string", default=None,
                      help="name of script to execute after failover is "
                           "complete and the utility has refreshed the "
                           "health report.")

    # Pedantic mode for failing if some inconsistencies are found
    parser.add_option("-p", "--pedantic", action="store_true", default=False,
                      dest="pedantic",
                      help="fail if some inconsistencies are found (e.g. "
                           "errant transactions on slaves).")

    # Add no keyboard input
    parser.add_option("--no-keyboard", action="store_true", default=False,
                      dest="no_keyboard",
                      help="start with no keyboard input support.")

    # Add option to run as daemon
    parser.add_option("--daemon", action="store", dest="daemon", default=None,
                      help="run on daemon mode. It can be start, stop, "
                           "restart or nodetach.",
                      type="choice",
                      choices=("start", "stop", "restart", "nodetach"))

    # Add pidfile for the daemon option
    parser.add_option("--pidfile", action="store", dest="pidfile",
                      type="string", default=None,
                      help="pidfile for running mysqlfailover as a daemon.")

    # Add report values for daemon mode
    parser.add_option("--report-values", action="store", dest="report_values",
                      type="string", default="health",
                      help="report values used in mysqlfailover running as a "
                           "daemon. It can be health, gtid or uuid. Multiple "
                           "values can be used separated by commas. The "
                           "default is health.")

    # Add verbosity mode
    add_verbosity(parser, False)

    # Replication user and password
    add_rpl_user(parser, None)

    # Now we process the rest of the arguments.
    opt, args = parser.parse_args()

    # Check slaves list
    if opt.daemon != "stop":
        check_server_lists(parser, opt.master, opt.slaves)

    # Check for errors
    if int(opt.interval) < 5:
        parser.error("The --interval option requires a value greater than or "
                     "equal to 5.")

    # The value for --timeout needs to be an integer > 0.
    try:
        if int(opt.timeout) <= 0:
            parser.error("The --timeout option requires a value greater "
                         "than 0.")
    except ValueError:
        parser.error("The --timeout option requires an integer value.")

    # if opt.master is None and opt.daemon and opt.daemon != "stop":
    if opt.master is None and opt.daemon != "stop":
        parser.error("You must specify a master to monitor.")

    if opt.slaves is None and opt.discover is None and opt.daemon != "stop":
        parser.error("You must supply a list of slaves or the "
                     "--discover-slaves-login option.")

    if opt.failover_mode == 'elect' and opt.candidates is None:
        parser.error("Failover mode = 'elect' requires at least one "
                     "candidate.")

    # Parse the master, slaves, and candidates connection parameters
    try:
        master_val, slaves_val, candidates_val = parse_topology_connections(
            opt)
    except UtilRplError:
        _, e, _ = sys.exc_info()
        print("ERROR: {0}".format(e.errmsg))
        sys.exit(1)

    # Check hostname alias
    for slave_val in slaves_val:
        if check_hostname_alias(master_val, slave_val):
            parser.error("The master and one of the slaves are the same "
                         "host and port.")
    for cand_val in candidates_val:
        if check_hostname_alias(master_val, cand_val):
            parser.error("The master and one of the candidates are the same "
                         "host and port.")

    # Create dictionary of options
    options = {
        'candidates': candidates_val,
        'ping': 3 if opt.ping is None else opt.ping,
        'verbosity': 0 if opt.verbosity is None else opt.verbosity,
        'before': opt.exec_before,
        'after': opt.exec_after,
        'fail_check': opt.exec_fail,
        'max_position': opt.max_position,
        'max_delay': opt.max_delay,
        'discover': opt.discover,
        'timeout': int(opt.timeout),
        'interval': opt.interval,
        'failover_mode': opt.failover_mode,
        'logging': opt.log_file is not None,
        'log_file': opt.log_file,
        'force': opt.force,
        'post_fail': opt.exec_post_fail,
        'rpl_user': opt.rpl_user,
        'pedantic': opt.pedantic,
        'no_keyboard': opt.no_keyboard,
        'daemon': opt.daemon,
        'pidfile': opt.pidfile,
        'report_values': opt.report_values,
        'script_threshold': opt.script_threshold,
    }

    # Purge log file of old data
    if opt.log_file is not None and not purge_log(opt.log_file, opt.log_age):
        parser.error("Error purging log file.")

    # Setup log file
    logging.basicConfig(filename=opt.log_file, level=logging.INFO,
                        format='%(asctime)s %(levelname)s %(message)s',
                        datefmt=_DATE_FORMAT)

    # Warn user about script threshold checking.
    if opt.script_threshold:
        print(SCRIPT_THRESHOLD_WARNING)

    # Check if the values specified for the --report-values option are valid.
    for report in opt.report_values.split(','):
        if report.lower() not in ("health", "gtid", "uuid"):
            parser.error("The value for the option --report-values is not "
                         "valid: '{0}', the values allowed are 'health', "
                         "'gitd' or 'uuid'".format(opt.report_values))

    # Check the daemon options
    if opt.daemon:
        # Check if a POSIX system
        if os.name != "posix":
            parser.error("Running mysqlfailover with --daemon is only "
                         "available for POSIX systems.")

        # Check the presence of --log
        if opt.daemon != "stop" and not opt.log_file:
            parser.error("The option --log is required when using --daemon.")

        # Test pidfile
        if opt.daemon != "nodetach":
            pidfile = opt.pidfile or "./failover_daemon.pid"
            pidfile = os.path.realpath(os.path.normpath(pidfile))
            if opt.daemon == "start":
                # Test if pidfile exists
                if os.path.exists(pidfile):
                    parser.error("pidfile {0} already exists. The daemon is "
                                 "already running?".format(pidfile))
                # Test if pidfile is writable
                try:
                    with open(pidfile, "w") as f:
                        f.write("{0}\n".format(0))
                    # Delete temporary pidfile
                    os.remove(pidfile)
                except IOError as err:
                    parser.error("Unable to write pidfile: {0}"
                                 "".format(err.strerror))
            else:
                # opt.daemon == stop/restart, test if pidfile is readable
                pid = None
                try:
                    if not os.path.exists(pidfile):
                        parser.error("pidfile {0} does not exist.".format(
                            pidfile))
                    with open(pidfile, "r") as f:
                        pid = int(f.read().strip())
                except IOError:
                    pid = None
                except ValueError:
                    pid = None
                # Test pid presence
                if not pid:
                    parser.error("Can not read pid from pidfile.")

    if opt.pidfile and not opt.daemon:
        parser.error("The option --daemon is required when using --pidfile.")

    try:
        rpl_cmds = RplCommands(master_val, slaves_val, options)
        if opt.daemon:
            rpl_cmds.auto_failover_as_daemon()
        else:
            rpl_cmds.auto_failover(opt.interval)
    except UtilError:
        _, e, _ = sys.exc_info()
        # log the error in case it was an usual exception
        logging.log(logging.CRITICAL, e.errmsg.strip(' '))
        print("ERROR: %s" % e.errmsg)
        sys.exit(1)
    except KeyboardInterrupt:
        sys.exit(0)

    sys.exit(0)



#
# Copyright (c) 2014 Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#
"""Methods and classes to store Fabric's log entries.
"""
import logging

from mysql.fabric import (
    persistence as _persistence,
    config as _config,
)

from mysql.fabric.utils import (
    get_time_from_timestamp,
)

_LOGGER = logging.getLogger(__name__)

_CREATE_FABRIC_LOG = (
    "CREATE TABLE log "
    "(subject VARCHAR(40) NOT NULL, "
    "reported TIMESTAMP(6) NOT NULL, "
    "reporter VARCHAR(64) NOT NULL, "
    "message TEXT, "
    "category int NOT NULL, "
    "type int NOT NULL, "
    "INDEX key_subject_reported (subject, reported), "
    "INDEX key_reporter (reporter), "
    "INDEX key_reported (reported), "
    "INDEX key_category (category), "
    "INDEX key_type (type))"
)

_CREATE_GROUP_VIEW = (
    "CREATE VIEW group_view AS SELECT subject as group_id, "
    "(SELECT COUNT(*) FROM log as in_log WHERE in_log.subject = group_id AND "
    "in_log.category = %s AND in_log.type = %s) as promote_count, "
    "(SELECT COUNT(*) FROM log as in_log WHERE in_log.subject = group_id AND "
    "in_log.category = %s AND in_log.type = %s) as demote_count "
    "FROM log as out_log WHERE out_log.category = %s GROUP BY out_log.subject"
)

_SELECT_GROUP_VIEW = (
    "SELECT * from group_view WHERE group_id LIKE %s"
)

_CREATE_EVENT_FABRIC_LOG = (
    "CREATE EVENT prune_log "
    "ON SCHEDULE EVERY %s SECOND "
    "DO DELETE FROM log WHERE "
    "TIMEDIFF(UTC_TIMESTAMP(), reported) > MAKETIME(%s,0,0)"
)

_CREATE_PROCEDURE_VIEW = (
    "CREATE VIEW proc_view AS SELECT out_log.subject as proc_name, "
    "(SELECT COUNT(*) FROM log as in_log WHERE in_log.subject = proc_name AND "
    "in_log.category = %s AND in_log.type = %s) as call_count, "
    "(SELECT COUNT(*) FROM log as in_log WHERE in_log.subject = proc_name AND "
    "in_log.category = %s AND in_log.type = %s) as abort_count "
    "FROM log as out_log WHERE out_log.category = %s GROUP BY out_log.subject"
)

_SELECT_PROCEDURE_VIEW = (
    "SELECT * FROM proc_view WHERE proc_name LIKE %s"
)

_INSERT_FABRIC_LOG = (
    "INSERT INTO log (subject, reported, reporter, "
    "message, category, type) VALUES(%s, %s, %s, %s, %s, %s)"
)


class MySQLFilter(logging.Filter):
    """Filter records that are not supposed to be written to the logging
    table.
    """
    def __init__(self, name):
        """Constructor for the MySQLFilter.
        """
        logging.Filter.__init__(self, name)

    def filter(self, record):
        """Filter a record that is not supposed to be written to the logging
        table. If the arguments to format the record does not contain subject,
        the record is ignored.
        """
        try:
            record.subject
        except AttributeError:
            return 0

        return 1


class MySQLHandler(logging.Handler, _persistence.Persistable):
    """Write information to the logging table if the appropriate options are
    set by the user. For example::

        logger.debug("Group %s has promote server %s to master.",
            group_id, server.uuid,
            extra={
                'subject' : group_id,
                'category' : MySQLHandler.GROUP,
                'type' : MySQLHandler.PROMOTE
            }
        )

    Users willing to write information to the logging table must specify the
    subject, category and type paramters as entries in a dictionary that is
    passed to the logging function through the extra parameter.

    The subject identifies the entity that the information is referring to,
    whereas the subject classifies the entity and the type defines what has
    been reported. In the aforementioned code, the subject is the group
    identifier, the category defines that information on a group is being
    reported and the type says that a new master is being promoted.

    The following categories are supported:

    * GROUP - Defines that information is about a group.
    * NODE - Defines that information is about the Fabric Node.
    * PROCEDURE - Defines that information is about a procedure.
    * SERVER - Defines that information is about a server.

    The following types are supported:

    * START - Defines that an entity has been started.
    * STOP - Defines that an entity has been stopped.
    * ABORT - Defines that an entity has been aborted. This is applicable
              to procedures.
    * PROMOTE - Defines that an entity has been promoted. This is applicable
                to a server or group.
    * DEMOTE - Defines that an entity has been demoted. This is applicable
               to a server or group.

    It is also possible to define the reporter and reported parameters. They
    define who has reported the information and when. If the reporter is not
    set, the name of the logger used to log the call is used. If the reported
    is not set, the time when the logging information was created is used.
    """
    MIN_PRUNE_TIME = 60
    PRUNE_TIME = _DEFAULT_PRUNE_TIME = 3600

    NODE = "NODE"
    PROCEDURE = "PROCEDURE"
    SERVER = "SERVER"
    GROUP = "GROUP"

    CATEGORIES = [ NODE, PROCEDURE, SERVER, GROUP ]

    START = "START"
    STOP = "STOP"
    ABORT = "ABORT"
    PROMOTE = "PROMOTE"
    DEMOTE = "DEMOTE"

    TYPES = [ START, STOP, ABORT, PROMOTE, DEMOTE ]

    @staticmethod
    def create(persister=None):
        """Create the objects(tables) that will store Fabric's logs.

        :param persister: Persister to persist the object to.
        :raises: DatabaseError If the table already exists.
        """
        persister.exec_stmt(_CREATE_FABRIC_LOG)
        persister.exec_stmt(
            _CREATE_EVENT_FABRIC_LOG % (MySQLHandler.PRUNE_TIME,
            MySQLHandler.PRUNE_TIME)
        )
        persister.exec_stmt(
            _CREATE_GROUP_VIEW % (
            MySQLHandler.idx_category(MySQLHandler.GROUP),
            MySQLHandler.idx_type(MySQLHandler.PROMOTE),
            MySQLHandler.idx_category(MySQLHandler.GROUP),
            MySQLHandler.idx_type(MySQLHandler.DEMOTE),
            MySQLHandler.idx_category(MySQLHandler.GROUP))
        )
        persister.exec_stmt(
            _CREATE_PROCEDURE_VIEW % (
            MySQLHandler.idx_category(MySQLHandler.PROCEDURE),
            MySQLHandler.idx_type(MySQLHandler.START),
            MySQLHandler.idx_category(MySQLHandler.PROCEDURE),
            MySQLHandler.idx_type(MySQLHandler.ABORT),
            MySQLHandler.idx_category(MySQLHandler.PROCEDURE))
        )

    @staticmethod
    def add(subject, reported, reporter, info, info_category,
            info_type, persister=None):
        """Add a Fabric's log entry.
        """
        persister.exec_stmt(_INSERT_FABRIC_LOG,
            {"params": (subject, reported, reporter, info, info_category,
            info_type)}
        )

    @staticmethod
    def group_view(group_id=None, persister=None):
        """Fetch information on a Group.

        The following information is returned:

            +----------------+----------------------+
            |    Column      |     Description      |
            +================+======================+
            |   group_id     | Group identification |
            +----------------+----------------------+
            |  promote_count | Many times a new     |
            |                | master was promoted  |
            +----------------+----------------------+
            |  demote_count  | Many times a master  |
            |                | was demoted          |
            +----------------+----------------------+

        :param group_id: Group identifier.
        :param persister: Persister to persist the object to.
        """
        like_pattern = None
        if not group_id:
            like_pattern = '%%'
        else:
            like_pattern = '%' + group_id + '%'

        res = persister.exec_stmt(_SELECT_GROUP_VIEW,
            {"raw" : False, "params": (like_pattern, )}
        )
        return res

    @staticmethod
    def procedure_view(procedure_name=None, persister=None):
        """Fetch information on a Procedure.

        The following information is returned:

            +--------------+----------------------+
            |    Column    |     Description      |
            +==============+======================+
            |   proc_name  |    Procedure name    |
            +--------------+----------------------+
            |   call_count | Number of times it   |
            |              | was called           |
            +--------------+----------------------+
            |   call_abort | Number of times it   |
            |              | has failed           |
            +--------------+----------------------+

        :param procedure_name: Procedure name.
        :param persister: Persister to persist the object to.
        """
        like_pattern = None
        if not procedure_name:
            like_pattern = '%%'
        else:
            like_pattern = '%' + procedure_name + '%'

        res = persister.exec_stmt(_SELECT_PROCEDURE_VIEW,
            {"raw" : False, "params":(like_pattern, )}
        )
        return res

    def __init__(self):
        """Constructor for MySQLHandler object.
        """
        logging.Handler.__init__(self)
        _persistence.Persistable.__init__(self)
        self.addFilter(MySQLFilter("MySQLHandler"))

    def emit(self, record):
        """Write an entry to the logging table if the extern key was provided as
        part of the extra dictionary and yields True.
        """
        try:
            info_subject = record.subject
            info_category = MySQLHandler.idx_category(record.category)
            info_type = MySQLHandler.idx_type(record.type)

            try:
                info_reporter = record.reporter
            except AttributeError:
                info_reporter = record.name

            try:
                info_reported = record.reported
            except AttributeError:
                info_reported = get_time_from_timestamp(record.created)

            MySQLHandler.add(
                info_subject, info_reported, info_reporter,
                self.format(record), info_category, info_type
            )
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)

    def flush(self):
        """Flush information to the logging table. There is no need to do
        anything here.
        """
        pass

    def close(self):
        """Close the handler.
        """
        logging.Handler.close(self)

    @staticmethod
    def idx_category(info_category):
        """Return an int representing a category.
        """
        return MySQLHandler.CATEGORIES.index(info_category)

    @staticmethod
    def idx_type(info_type):
        """Return an int representing a type.
        """
        return MySQLHandler.TYPES.index(info_type)

def configure(config):
    """Set configuration values.
    """
    try:
        prune_time = int(config.get("statistics", "prune_time"))
        if prune_time < MySQLHandler.MIN_PRUNE_TIME:
            _LOGGER.warning(
                "Prune entries cannot be lower than %s.",
                MySQLHandler.MIN_PRUNE_TIME
            )
            prune_time = MySQLHandler.MIN_PRUNE_TIME
        MySQLHandler.PRUNE_TIME = int(prune_time)
    except (_config.NoOptionError, _config.NoSectionError, ValueError):
        pass



#
# Copyright (c) 2014 Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#
"""Methods and classes to report servers' errors.
"""
import logging

from mysql.fabric import (
    persistence as _persistence,
    config as _config,
)

from mysql.fabric.utils import (
   get_time,
)

_LOGGER = logging.getLogger(__name__)

class ErrorLog(_persistence.Persistable):
    """Errors reported by an entity that suspects that a server has failed
    are processed and stored in a table (i.e. error_log) for future
    analysis. Each entry has the following fields:

    * erver_uuid - Server's UUID.
    * reported - When the error has been reported.
    * reporter - Who has reported the error.
    * error - Error description.

    See also:

    * :class:`~mysql.fabric.server.Group`.
    * :class:`~mysql.fabric.services.threat.ReportError`.
    * :class:`~mysql.fabric.services.threat.ReportFailure`.
    """

    _MIN_PRUNE_TIME = 60
    _PRUNE_TIME = _DEFAULT_PRUNE_TIME = 3600

    CREATE_SERVER_ERROR_LOG = (
        "CREATE TABLE error_log "
        "(server_uuid VARCHAR(40) NOT NULL, "
        "reported TIMESTAMP(6) NOT NULL, "
        "reporter VARCHAR(64) NOT NULL, "
        "error TEXT, "
        "INDEX key_server_uuid_reported (server_uuid, reported), "
        "INDEX key_reporter (reporter))"
    )

    #SQL Statement for creating event used to prune error logs that are
    #older than PRUNE_TIME.
    CREATE_EVENT_ERROR_LOG = (
        "CREATE EVENT prune_error_log "
        "ON SCHEDULE EVERY %s SECOND "
        "DO DELETE FROM error_log WHERE "
        "TIMEDIFF(UTC_TIMESTAMP(), reported) > MAKETIME(%s,0,0)"
    )

    ADD_FOREIGN_KEY_CONSTRAINT_SERVER_UUID = (
        "ALTER TABLE error_log ADD CONSTRAINT "
        "fk_server_uuid_error_log FOREIGN KEY(server_uuid) "
        "REFERENCES servers(server_uuid)"
    )

    QUERY_SERVER_ERROR_LOG = (
        "SELECT reported, reporter FROM error_log WHERE "
        "server_uuid = %s AND reported >= %s ORDER BY reported ASC"
    )

    REMOVE_SERVER_ERROR_LOG = (
        "DELETE FROM error_log WHERE server_uuid = %s"
    )

    INSERT_SERVER_ERROR_LOG = (
        "INSERT INTO error_log (server_uuid, reported, reporter, "
        "error) VALUES(%s, %s, %s, %s)"
    )

    @staticmethod
    def create(persister=None):
        """Create the objects(tables) that will store server's errors.

        :param persister: Persister to persist the object to.
        :raises: DatabaseError If the table already exists.
        """
        persister.exec_stmt(ErrorLog.CREATE_SERVER_ERROR_LOG)
        persister.exec_stmt(
            ErrorLog.CREATE_EVENT_ERROR_LOG % (ErrorLog._PRUNE_TIME,
            ErrorLog._PRUNE_TIME, )
        )

    @staticmethod
    def add_constraints(persister=None):
        """Add the constraints to the error_log table.

        :param persister: The DB server that can be used to access the
                          state store.
        """
        persister.exec_stmt(
                ErrorLog.ADD_FOREIGN_KEY_CONSTRAINT_SERVER_UUID)

    @staticmethod
    def add(server, reported, reporter, error, persister=None):
        """Add a server's error log.

        :param server: Server where the error has happened.
        :param reported: When the error has happened.
        :param reporter: Entity that has reported the error.
        :param error: Error reported.
        """
        from mysql.fabric.server import MySQLServer
        assert(isinstance(server, MySQLServer))
        persister.exec_stmt(ErrorLog.INSERT_SERVER_ERROR_LOG,
            {"params": (str(server.uuid), reported, reporter, error)})

    @staticmethod
    def remove(server, persister=None):
        """Remove server's error logs from the state store.

        :param server: A reference to a server.
        :param persister: Persister to persist the object to.
        """
        from mysql.fabric.server import MySQLServer
        assert(isinstance(server, MySQLServer))
        persister.exec_stmt(
            ErrorLog.REMOVE_SERVER_ERROR_LOG,
            {"params": (str(server.uuid), )}
            )

    @staticmethod
    def fetch(server, interval, now=None, persister=None):
        """Return a ErrorLog object corresponding to the
        server.

        :param server: Server whose error has been reported.
        :param interval: Interval of interest.
        :param now: Consider from `now` until `now` - `interval`.
        :param persister: Persister to persist the object to.
        :return: ErrorLog object.
        """
        from mysql.fabric.server import MySQLServer
        assert(isinstance(server, MySQLServer))

        now = now or get_time()
        (whens, reporters) = ErrorLog.compute(
            server.uuid, interval, now
        )
        return ErrorLog(server, interval, now, whens, reporters)

    @staticmethod
    def compute(uuid, interval, now, persister=None):
        """Compute information to build a ErrorLog object.

        :param server: Server's UUID whose error has been reported.
        :param interval: Interval of interest.
        :param now: Consider from `now` until `now` - `interval`.
        :param persister: Persister to persist the object to.
        :return: ErrorLog object.
        """

        cur = persister.exec_stmt(
            ErrorLog.QUERY_SERVER_ERROR_LOG,
            {"fetch" : False, "raw" : False,
             "params": (str(uuid), now - interval)
            }
        )

        whens = []
        reporters = []
        for row in cur.fetchall():
            when, reporter = row
            whens.append(when)
            reporters.append(reporter)

        return (whens, reporters)

    def __init__(self, server, interval, now, whens, reporters):
        """Constructor for ErrorLog object.
        """
        from mysql.fabric.server import MySQLServer
        assert(isinstance(server, MySQLServer))
        assert(len(whens) == len(reporters))

        super(ErrorLog, self).__init__()

        self.__server_uuid = server.uuid
        self.__now = now
        self.__interval = interval
        self.__whens = whens or []
        self.__reporters = reporters or []

    @property
    def server_uuid(self):
        """Return the associated server's UUID.
        """
        return self.__server_uuid

    @property
    def whens(self):
        """Return when errors were reported.
        interval.
        """
        return self.__whens

    @property
    def reporters(self):
        """Return who reported errors.
        """
        return self.__reporters

    @property
    def now(self):
        """Base time to analyze errors.
        """
        return self.__now

    @property
    def interval(self):
        """Interval upon which errors are analyzed.
        """
        return self.__interval

    def refresh(self):
        """Refresh set of errors.
        """
        (whens, reporters) = ErrorLog.compute(
            self.__server_uuid, self.__interval, self.__now
        )
        self.__whens = whens
        self.__reporters = reporters

    def is_unstable(self, n_notifications, n_reporters, filter_reporter=None):
        """Check whether a server is unstable or not.

        :param n_notifications: Number of notifications threshold.
        :param n_reporters: Number of reporters threshold.
        :param filter_reporter: Filter errors by reporters before checking
                                whether a server is unstable or not.
        :return: Whether the server can be considered faulty
                 or not. The outcome depends on the parameters.
        """
        assert(
            filter_reporter is None or isinstance(filter_reporter, tuple) or \
            isinstance(filter_reporter, list)
        )

        reporters = self.__reporters
        if filter_reporter is not None:
            reporters = [
                reporter for reporter in self.__reporters \
                if reporter in filter_reporter
            ]

        nreporters = len(set(reporters))
        notifications = len(reporters)
        if nreporters >= n_reporters and notifications >= n_notifications:
            return True

        return False

def configure(config):
    """Set configuration values.
    """
    try:
        prune_time = int(config.get("failure_tracking", "prune_time"))
        if prune_time < ErrorLog._MIN_PRUNE_TIME:
            _LOGGER.warning(
                "Prune entries cannot be lower than %s.",
                ErrorLog._MIN_PRUNE_TIME
            )
            prune_time = ErrorLog._MIN_PRUNE_TIME
        ErrorLog._PRUNE_TIME = int(prune_time)
    except (_config.NoOptionError, _config.NoSectionError, ValueError):
        pass



#
# Copyright (c) 2013,2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""Module contains the abstract classes for implementing backup and
restoring a server from a backup. The module also contains the concrete
class for implementing the concrete class for doing backup using
mysqldump.
"""

from abc import ABCMeta, abstractmethod

import logging
import shlex
import subprocess

from subprocess import (
    Popen,
    PIPE,
)

from mysql.fabric import (
    errors as _errors,
    server_utils as _server_utils,
)

from urlparse import urlparse

from mysql.fabric.server import MySQLServer

_LOGGER = logging.getLogger(__name__)

class BackupImage(object):
    """Class that represents a backup image to which the output
    of a backup method is directed.

    :param uri: The URI to the destination on which the backup image needs
                       to be stored.
    """
    def __init__(self,  uri):
        """The constructor for initializing the backup image.
        """

        self.__uri = uri
        if self.__uri is not None:
            parsed = urlparse(self.__uri)
            self.__scheme = parsed.scheme
            self.__netloc = parsed.netloc
            self.__path = parsed.path
            self.__params = parsed.params
            self.__query = parsed.query
            self.__fragment = parsed.fragment
            self.__username = parsed.username
            self.__password = parsed.password
            self.__hostname = parsed.hostname
            self.__port = parsed.port

    @property
    def uri(self):
        """Return the uri.
        """
        return self.__uri

    @property
    def scheme(self):
        """Return the type of the URI like ftp, http, file etc.
        """
        return self.__scheme

    @property
    def netloc(self):
        """Return the network location in the URI.
        """
        return self.__netloc

    @property
    def path(self):
        """Return the path to the file represented by the URI.
        """
        return self.__path

    @property
    def params(self):
        """Return the parameters in the URI.
        """
        return self.__params

    @property
    def query(self):
        """Return the query in the backup image URI.
        """
        return self.__query

    @property
    def fragment(self):
        """Return the fragment in the backup image URI.
        """
        return self.__fragment

    @property
    def username(self):
        """Return the username in the backup image URI.
        """
        return self.__username

    @property
    def password(self):
        """Return the password in the backup image URI.
        """
        return self.__password

    @property
    def hostname(self):
        """Return the hostname in the backup image URI.
        """
        return self.__hostname

    @property
    def port(self):
        """Return the port number in the backup image URI.
        """
        return self.__port

class BackupMethod(object):
    """Abstract class that represents the interface methods that need to be
    implemented by a class that encapsulates a MySQL backup and restore
    method.
    """
    __metaclass__ = ABCMeta

    @staticmethod
    def backup(server):
        """Perform the backup.

        :param server: The server that needs to be backed up.
        """
        pass

    @staticmethod
    def restore(server,  image):
        """Restore the backup from the image to the server.

        :param server: The server on which the backup needs to be restored.
        :param image: The image that needs to be restored.
        """
        pass

    @staticmethod
    def copyBackup(image):
        """Will be used in cases when the backup needs to be taken on a source
        and will be copied to the destination machine.

        :param image: BackupImage object containning the location where the
                      backup needs to be copied to.
        """
        pass

class MySQLDump(BackupMethod):
    """Class that implements the BackupMethod abstract interface
    using MySQLDump.
    """

    MYSQL_DEFAULT_PORT = 3306

    @staticmethod
    def dump_to_log(message, error_lines=None):
        """Write the error lines to the log.

        :param message: The message signifying the start of the below log.
        :param error_lines: The error lines that were output by the command.
        """
        if error_lines:
            _LOGGER.debug(message + ": " + error_lines)

    @staticmethod
    def backup(server, config_file, mysqldump_binary=None):
        """Perform the backup using mysqldump.

        The backup results in creation a .sql file on the FABRIC server,
        this method needs to be optimized going forward. But for now
        this will suffice.

        :param server: The MySQLServer that needs to be backed up.
        :param config_file: The complete path to the fabric configuration file.
        :param mysqldump_binary: The fully qualified mysqldump binary.
        """
        assert (isinstance(server, MySQLServer))

        #Extract the host and the port from the server address.
        host = None
        port = None
        if server.address is not None:
            host, port = _server_utils.split_host_port(
                            server.address,
                            MySQLDump.MYSQL_DEFAULT_PORT
                         )

        #Form the name of the destination .sql file from the name of the
        #server host and the port number that is being backed up.
        destination = "MySQL_{HOST}_{PORT}.sql".format(
                                                    HOST=host,
                                                    PORT=port)

        mysqldump_command = shlex.split(mysqldump_binary)

        #Setup the MYSQLDump command that is used to backup the server.
        #Append the password parameter if the password is not None.
        mysqldump_command.extend([
            "--defaults-extra-file="+config_file,
            "--all-databases", "--single-transaction",
            "--add-drop-table",
            "--triggers", "--routines",
            "--events",
            "--protocol=tcp",
            "-h" + str(host),
            "-P" + str(port),
            "-u" + str(server.user)])

        #Run the backup command
        with open(destination,"w") as fd_file:
            p = Popen(mysqldump_command, stdout=fd_file, stderr=PIPE)
            _, error_lines = p.communicate()
            if p.returncode:
                MySQLDump.dump_to_log(
                    "Error while taking backup using MySQLDump\n",
                    error_lines
                )
                raise _errors.BackupError(
                    "Error while taking backup using MySQLDump\n, %s",
                    error_lines
                )

        #Return the backup image containing the location of the .sql file.
        return BackupImage(destination)

    @staticmethod
    def restore_server(host,  port, user, passwd,
                        image, config_file, mysqlclient_binary):
        """Restore the backup from the image to a server outside the Fabric
        Farm.

        In the current implementation the restore works by restoring the
        .sql file created on the FABRIC server. This will be optimized in future
        implementation. But for the current implementation this suffices.

        :param host: The host name of the server on which the backup needs
                    to be restored into.
        :param port: The port number of the server on which the backup needs
                    to be restored into
        :param user: The user name used for accessing the server.
        :param passwd: The password used for accessing the server.
        :param image: The image that needs to be restored.
        :param config_file: The complete path to the fabric configuration file.
        :param mysqlclient_binary: The fully qualified mysqlclient binary.
        """

        mysqlclient_command = shlex.split(mysqlclient_binary)

        #Use the mysql client for the restore. Append the -p option with
        #the password if there is a non-empty password.
        mysqlclient_command.extend([
            "--defaults-extra-file="+config_file,
            "-h" + str(host),  "-P" + str(port),
            "--protocol=tcp",
            "-u" + str(user)])

        #Fire the mysql client for the restore using the input image as
        #the restore source.
        p = Popen(mysqlclient_command, stdin=PIPE, stderr=PIPE)
        p.stdin.write("source %s\n" % (image.path,))
        _, error_lines = p.communicate()
        if p.returncode:
            MySQLDump.dump_to_log(
                "Error while restoring the backup using "\
                "the mysql client\n",
                error_lines
            )
            raise _errors.BackupError(
                "Error while restoring the backup using "\
                "the mysql client\n, %s",
                error_lines
            )

    @staticmethod
    def restore_fabric_server(server,  image, config_file, mysqlclient_binary):
        """Restore the backup from the image to a server within the
        fabric farm and managed by the Fabric server.

        In the current implementation the restore works by restoring the
        .sql file created on the FABRIC server. This will be optimized in future
        implementation. But for the current implementation this suffices.

        :param server: The server on which the backup needs to be restored.
        :param image: The image that needs to be restored.
        :param config_file: The complete path to the fabric configuration file.
        :param mysqlclient_binary: The fully qualified mysqlclient binary.
        """

        assert (isinstance(server, MySQLServer))
        assert (image is None or isinstance(image, BackupImage))

        #Extract the host and the port from the server address.
        host = None
        port = None
        if server.address is not None:
            host, port = _server_utils.split_host_port(
                                                server.address,
                                                MySQLDump.MYSQL_DEFAULT_PORT
                                                )
        MySQLDump.restore_server(host, port, server.user, server.passwd,
                image, config_file, mysqlclient_binary)

    @abstractmethod
    def copyBackup(image):
        """Currently the MySQLDump based backup method works by backing
        up on the FABRIC server and restoring from there. This method is not
        required for the current implemention of MySQDump based backup.
        """
        pass



#
# Copyright (c) 2013,2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""This module provides the necessary interfaces for performing administrative
tasks on replication.
"""
import re
import logging
import uuid as _uuid

import mysql.fabric.services.utils as _utils

from  mysql.fabric import (
    events as _events,
    group_replication as _group_replication,
    server as _server,
    replication as _replication,
    errors as _errors,
)

from mysql.fabric.command import (
    ProcedureGroup,
)

from mysql.fabric.services.server import (
    _retrieve_server
)

_LOGGER = logging.getLogger(__name__)

# Find out which operation should be executed.
DEFINE_HA_OPERATION = _events.Event()
# Find a slave that was not failing to keep with the master's pace.
FIND_CANDIDATE_FAIL = _events.Event("FAIL_OVER")
# Check if the candidate is properly configured to become a master.
CHECK_CANDIDATE_FAIL = _events.Event()
# Wait until all slaves or a candidate process the relay log.
WAIT_SLAVE_FAIL = _events.Event()
# Find a slave that is not failing to keep with the master's pace.
FIND_CANDIDATE_SWITCH = _events.Event()
# Check if the candidate is properly configured to become a master.
CHECK_CANDIDATE_SWITCH = _events.Event()
# Block any write access to the master.
BLOCK_WRITE_SWITCH = _events.Event()
# Wait until all slaves synchronize with the master.
WAIT_SLAVES_SWITCH = _events.Event()
# Enable the new master by making slaves point to it.
CHANGE_TO_CANDIDATE = _events.Event()
class PromoteMaster(ProcedureGroup):
    """Promote a server into master.

    If users just want to update the state store and skip provisioning steps
    such as configuring replication, the update_only parameter must be set to
    true. Otherwise, the following happens.

    If the master within a group fails, a new master is either automatically
    or manually selected among the slaves in the group. The process of
    selecting and setting up a new master after detecting that the current
    master failed is known as failover.

    It is also possible to switch to a new master when the current one is still
    alive and kicking. The process is known as switchover and may be used, for
    example, when one wants to take the current master off-line for
    maintenance.

    If a slave is not provided, the best candidate to become the new master is
    found. Any candidate must have the binary log enabled, should
    have logged the updates executed through the SQL Thread and both
    candidate and master must belong to the same group. The smaller the lag
    between slave and the master the better. So the candidate which satisfies
    the requirements and has the smaller lag is chosen to become the new
    master.

    In the failover operation, after choosing a candidate, one makes the slaves
    point to the new master and updates the state store setting the new master.

    In the switchover operation, after choosing a candidate, any write access
    to the current master is disabled and the slaves are synchronized with it.
    Failures during the synchronization that do not involve the candidate slave
    are ignored. Then slaves are stopped and configured to point to the new
    master and the state store is updated setting the new master.
    """
    group_name = "group"
    command_name = "promote"

    def execute(self, group_id, slave_id=None, update_only=False,
                synchronous=True):
        """Promote a new master.

        :param uuid: Group's id.
        :param slave_id: Candidate's UUID or HOST:PORT.
        :param update_only: Only update the state store and skip provisioning.
        :param synchronous: Whether one should wait until the execution finishes
                            or not.

        In what follows, one will find a figure that depicts the sequence of
        event that happen during a promote operation. The figure is split in
        two pieces and names are abbreviated in order to ease presentation:

        .. seqdiag::

          diagram {
            activation = none;
            === Schedule "find_candidate" ===
            fail_over --> executor [ label = "schedule(find_candidate)" ];
            fail_over <-- executor;
            === Execute "find_candidate" and schedule "check_candidate" ===
            executor -> find_candidate [ label = "execute(find_candidate)" ];
            find_candidate --> executor [ label = "schedule(check_candidate)" ];
            find_candidate <-- executor;
            executor <- find_candidate;
            === Execute "check_candidate" and schedule "wait_slave" ===
            executor -> check_candidate [ label = "execute(check_candidate)" ];
            check_candidate --> executor [ label = "schedule(wait_slave)" ];
            check_candidate <-- executor;
            executor <- check_candidate;
            === Continue in the next diagram ===
          }

        .. seqdiag::

          diagram {
            activation = none;
            edge_length = 300;
            === Continuation from previous diagram ===
            === Execute "wait_slaves" and schedule "change_to_candidate" ===
            executor -> wait_slave [ label = "execute(wait_slave)" ];
            wait_slave --> executor [ label = "schedule(change_to_candidate)" ];
            wait_slave <-- executor;
            executor <- wait_slave;
            === Execute "change_to_candidate" ===
            executor -> change_to_candidate [ label = "execute(change_to_candidate)" ];
            change_to_candidate <- executor;
          }

        In what follows, one will find a figure that depicts the sequence of
        events that happen during the switchover operation. The figure is split
        in two pieces and names are abreviated in order to ease presentation:

        .. seqdiag::

          diagram {
            activation = none;
            === Schedule "find_candidate" ===
            switch_over --> executor [ label = "schedule(find_candidate)" ];
            switch_over <-- executor;
            === Execute "find_candidate" and schedule "check_candidate" ===
            executor -> find_candidate [ label = "execute(find_candidate)" ];
            find_candidate --> executor [ label = "schedule(check_candidate)" ];
            find_candidate <-- executor;
            executor <- find_candidate;
            === Execute "check_candidate" and schedule "block_write" ===
            executor -> check_candidate [ label = "execute(check_candidate)" ];
            check_candidate --> executor [ label = "schedule(block_write)" ];
            check_candidate <-- executor;
            executor <- check_candidate;
            === Execute "block_write" and schedule "wait_slaves" ===
            executor -> block_write [ label = "execute(block_write)" ];
            block_write --> executor [ label = "schedule(wait_slaves)" ];
            block_write <-- executor;
            executor <- block_write;
            === Continue in the next diagram ===
          }

        .. seqdiag::

          diagram {
            activation = none;
            edge_length = 350;
            === Continuation from previous diagram ===
            === Execute "wait_slaves_catch" and schedule "change_to_candidate" ===
            executor -> wait_slaves [ label = "execute(wait_slaves)" ];
            wait_slaves --> executor [ label = "schedule(change_to_candidate)" ];
            wait_slaves <-- executor;
            executor <- wait_slaves;
            === Execute "change_to_candidate" ===
            executor -> change_to_candidate [ label = "execute(change_to_candidate)" ];
            executor <- change_to_candidate;
          }
        """
        procedures = _events.trigger(
            DEFINE_HA_OPERATION, self.get_lockable_objects(), group_id,
            slave_id, update_only
        )
        return self.wait_for_procedures(procedures, synchronous)

@_events.on_event(DEFINE_HA_OPERATION)
def _define_ha_operation(group_id, slave_id, update_only):
    """Define which operation must be called based on the master's status
    and whether the candidate slave is provided or not.
    """
    fail_over = True

    group = _server.Group.fetch(group_id)
    if not group:
        raise _errors.GroupError("Group (%s) does not exist." % (group_id, ))

    if update_only and not slave_id:
        raise _errors.ServerError(
            "The new master must be specified through --slave-uuid if "
            "--update-only is set."
        )

    if group.master:
        master = _server.MySQLServer.fetch(group.master)
        if master.status != _server.MySQLServer.FAULTY:
            if update_only:
                _do_block_write_master(group_id, str(group.master), update_only)
            fail_over = False

    if update_only:
        # Check whether the server is registered or not.
        _retrieve_server(slave_id, group_id)
        _change_to_candidate(group_id, slave_id, update_only)
        return

    if fail_over:
        if not slave_id:
            _events.trigger_within_procedure(FIND_CANDIDATE_FAIL, group_id)
        else:
            _events.trigger_within_procedure(CHECK_CANDIDATE_FAIL, group_id,
                                             slave_id
            )
    else:
        if not slave_id:
            _events.trigger_within_procedure(FIND_CANDIDATE_SWITCH, group_id)
        else:
            _events.trigger_within_procedure(CHECK_CANDIDATE_SWITCH, group_id,
                                             slave_id
            )

# Block any write access to the master.
BLOCK_WRITE_DEMOTE = _events.Event()
# Wait until all slaves synchronize with the master.
WAIT_SLAVES_DEMOTE = _events.Event()
class DemoteMaster(ProcedureGroup):
    """Demote the current master if there is one.

    If users just want to update the state store and skip provisioning steps
    such as configuring replication, the update_only parameter must be set to
    true. Otherwise any write access to the master is blocked, slaves are
    synchronized with the master, stopped and their replication configuration
    reset. Note that no slave is promoted as master.
    """
    group_name = "group"
    command_name = "demote"

    def execute(self, group_id, update_only=False, synchronous=True):
        """Demote the current master if there is one.

        :param uuid: Group's id.
        :param update_only: Only update the state store and skip provisioning.
        :param synchronous: Whether one should wait until the execution finishes
                            or not.

        In what follows, one will find a figure that depicts the sequence of
        event that happen during the demote operation. To ease the presentation
        some names are abbreivated:

        .. seqdiag::

          diagram {
            activation = none;
            === Schedule "block_write" ===
            demote --> executor [ label = "schedule(block_write)" ];
            demote <-- executor;
            === Execute "block_write" and schedule "wait_slaves" ===
            executor -> block_write [ label = "execute(block_write)" ];
            block_write --> executor [ label = "schedule(wait_slaves)" ];
            block_write <-- executor;
            executor <- block_write;
            === Execute "wait_slaves" ===
            executor -> wait_slaves [ label = "execute(wait_slaves)" ];
            wait_slaves --> executor;
            wait_slaves <-- executor;
            executor <- wait_slaves;
          }
        """
        procedures = _events.trigger(
            BLOCK_WRITE_DEMOTE, self.get_lockable_objects(), group_id,
            update_only
        )
        return self.wait_for_procedures(procedures, synchronous)

@_events.on_event(FIND_CANDIDATE_SWITCH)
def _find_candidate_switch(group_id):
    """Find the best slave to replace the current master.
    """
    slave_uuid = _do_find_candidate(group_id, FIND_CANDIDATE_SWITCH)
    _events.trigger_within_procedure(CHECK_CANDIDATE_SWITCH, group_id,
                                     slave_uuid)

def _do_find_candidate(group_id, event):
    """Find the best candidate in a group that may be used to replace the
    current master if there is any.

    It chooses the slave that has processed more transactions and may become a
    master, e.g. has the binary log enabled. This function does not consider
    purged transactions and delays in the slave while picking up a slave.

    :param group_id: Group's id from where a candidate will be chosen.
    :return: Return the uuid of the best candidate to become a master in the
             group.
    """
    forbidden_status = (_server.MySQLServer.FAULTY, _server.MySQLServer.SPARE)
    group = _server.Group.fetch(group_id)

    master_uuid = None
    if group.master:
        master_uuid = str(group.master)

    chosen_uuid = None
    chosen_gtid_status = None
    for candidate in group.servers():
        if master_uuid != str(candidate.uuid) and \
            candidate.status not in forbidden_status:
            try:
                candidate.connect()
                gtid_status = candidate.get_gtid_status()
                master_issues = \
                    _replication.check_master_issues(candidate)
                if event == FIND_CANDIDATE_SWITCH:
                    slave_issues = \
                        _replication.check_slave_issues(candidate)
                else:
                    slave_issues = {}
                has_valid_master = (master_uuid is None or \
                    _replication.slave_has_master(candidate) == master_uuid)
                can_become_master = False
                if chosen_gtid_status:
                    n_trans = 0
                    try:
                        n_trans = _replication.get_slave_num_gtid_behind(
                            candidate, chosen_gtid_status
                            )
                    except _errors.InvalidGtidError:
                        pass
                    if n_trans == 0 and not master_issues and \
                        has_valid_master and not slave_issues:
                        chosen_gtid_status = gtid_status
                        chosen_uuid = str(candidate.uuid)
                        can_become_master = True
                elif not master_issues and has_valid_master and \
                    not slave_issues:
                    chosen_gtid_status = gtid_status
                    chosen_uuid = str(candidate.uuid)
                    can_become_master = True
                if not can_become_master:
                    _LOGGER.warning(
                        "Candidate (%s) cannot become a master due to the "
                        "following reasons: issues to become a "
                        "master (%s), prerequistes as a slave (%s), valid "
                        "master (%s).", candidate.uuid, master_issues,
                        slave_issues, has_valid_master
                        )
            except _errors.DatabaseError as error:
                _LOGGER.warning(
                    "Error accessing candidate (%s).", candidate.uuid,
                    exc_info=error
                )

    if not chosen_uuid:
        raise _errors.GroupError(
            "There is no valid candidate that can be automatically "
            "chosen in group (%s). Please, choose one manually." %
            (group_id, )
            )
    return chosen_uuid

@_events.on_event(CHECK_CANDIDATE_SWITCH)
def _check_candidate_switch(group_id, slave_id):
    """Check if the candidate has all the features to become the new
    master.
    """
    allowed_status = (_server.MySQLServer.SECONDARY, _server.MySQLServer.SPARE)
    group = _server.Group.fetch(group_id)

    if not group.master:
        raise _errors.GroupError(
            "Group (%s) does not contain a valid "
            "master. Please, run a promote or failover." % (group_id, )
        )

    slave = _retrieve_server(slave_id, group_id)
    slave.connect()

    if group.master == slave.uuid:
        raise _errors.ServerError(
            "Candidate slave (%s) is already master." % (slave_id, )
            )

    master_issues = _replication.check_master_issues(slave)
    if master_issues:
        raise _errors.ServerError(
            "Server (%s) is not a valid candidate slave "
            "due to the following reason(s): (%s)." %
            (slave.uuid, master_issues)
            )

    slave_issues = _replication.check_slave_issues(slave)
    if slave_issues:
        raise _errors.ServerError(
            "Server (%s) is not a valid candidate slave "
            "due to the following reason: (%s)." %
            (slave.uuid, slave_issues)
            )

    master_uuid = _replication.slave_has_master(slave)
    if master_uuid is None or group.master != _uuid.UUID(master_uuid):
        raise _errors.GroupError(
            "The group's master (%s) is different from the candidate's "
            "master (%s)." % (group.master, master_uuid)
            )

    if slave.status not in allowed_status:
        raise _errors.ServerError("Server (%s) is faulty." % (slave_id, ))

    _events.trigger_within_procedure(
        BLOCK_WRITE_SWITCH, group_id, master_uuid, str(slave.uuid)
        )

@_events.on_event(BLOCK_WRITE_SWITCH)
def _block_write_switch(group_id, master_uuid, slave_uuid):
    """Block and disable write access to the current master.
    """
    _do_block_write_master(group_id, master_uuid)
    _events.trigger_within_procedure(WAIT_SLAVES_SWITCH, group_id,
        master_uuid, slave_uuid
        )

def _do_block_write_master(group_id, master_uuid, update_only=False):
    """Block and disable write access to the current master.

    Note that connections are not killed and blocking the master
    may take some time.
    """
    master = _server.MySQLServer.fetch(_uuid.UUID(master_uuid))
    assert(master.status == _server.MySQLServer.PRIMARY)
    master.mode = _server.MySQLServer.READ_ONLY
    master.status = _server.MySQLServer.SECONDARY

    if not update_only:
        master.connect()
        _utils.set_read_only(master, True)

    # Temporarily unset the master in this group.
    group = _server.Group.fetch(group_id)
    _set_group_master_replication(group, None)

    # At the end, we notify that a server was demoted.
    _events.trigger("SERVER_DEMOTED", set([group_id]),
        group_id, str(master.uuid)
    )

@_events.on_event(WAIT_SLAVES_SWITCH)
def _wait_slaves_switch(group_id, master_uuid, slave_uuid):
    """Synchronize candidate with master and also all the other slaves.

    Note that this can be optimized as one may determine the set of
    slaves that must be synchronized with the master.
    """
    master = _server.MySQLServer.fetch(_uuid.UUID(master_uuid))
    master.connect()
    slave = _server.MySQLServer.fetch(_uuid.UUID(slave_uuid))
    slave.connect()

    _utils.synchronize(slave, master)
    _do_wait_slaves_catch(group_id, master, [slave_uuid])

    _events.trigger_within_procedure(CHANGE_TO_CANDIDATE, group_id, slave_uuid)

def _do_wait_slaves_catch(group_id, master, skip_servers=None):
    """Synchronize slaves with master.
    """
    skip_servers = skip_servers or []
    skip_servers.append(str(master.uuid))

    group = _server.Group.fetch(group_id)
    for server in group.servers():
        if str(server.uuid) not in skip_servers:
            try:
                server.connect()
                used_master_uuid = _replication.slave_has_master(server)
                if  str(master.uuid) == used_master_uuid:
                    _utils.synchronize(server, master)
                else:
                    _LOGGER.debug("Slave (%s) has a different master "
                        "from group (%s).", server.uuid, group_id)
            except _errors.DatabaseError as error:
                _LOGGER.debug(
                    "Error synchronizing slave (%s).", server.uuid,
                    exc_info=error
                )

@_events.on_event(CHANGE_TO_CANDIDATE)
def _change_to_candidate(group_id, master_uuid, update_only=False):
    """Switch to candidate slave.
    """
    forbidden_status = (_server.MySQLServer.FAULTY, )
    master = _server.MySQLServer.fetch(_uuid.UUID(master_uuid))
    master.mode = _server.MySQLServer.READ_WRITE
    master.status = _server.MySQLServer.PRIMARY

    if not update_only:
        # Prepare the server to be the master
        master.connect()
        _utils.reset_slave(master)
        _utils.set_read_only(master, False)

    group = _server.Group.fetch(group_id)
    _set_group_master_replication(group, master.uuid, update_only)

    if not update_only:
        # Make slaves point to the master.
        for server in group.servers():
            if server.uuid != _uuid.UUID(master_uuid) and \
                server.status not in forbidden_status:
                try:
                    server.connect()
                    _utils.switch_master(server, master)
                except _errors.DatabaseError as error:
                    _LOGGER.debug(
                        "Error configuring slave (%s).", server.uuid,
                        exc_info=error
                    )

    # At the end, we notify that a server was promoted.
    _events.trigger("SERVER_PROMOTED", set([group_id]),
        group_id, master_uuid
    )

@_events.on_event(FIND_CANDIDATE_FAIL)
def _find_candidate_fail(group_id):
    """Find the best candidate to replace the failed master.
    """
    slave_uuid = _do_find_candidate(group_id, FIND_CANDIDATE_FAIL)
    _events.trigger_within_procedure(CHECK_CANDIDATE_FAIL, group_id,
                                     slave_uuid)

@_events.on_event(CHECK_CANDIDATE_FAIL)
def _check_candidate_fail(group_id, slave_id):
    """Check if the candidate has all the prerequisites to become the new
    master.
    """
    allowed_status = (_server.MySQLServer.SECONDARY, _server.MySQLServer.SPARE)
    group = _server.Group.fetch(group_id)

    slave = _retrieve_server(slave_id, group_id)
    slave.connect()

    if group.master == slave.uuid:
        raise _errors.ServerError(
            "Candidate slave (%s) is already master." % (slave_id, )
            )

    master_issues = _replication.check_master_issues(slave)
    if master_issues:
        raise _errors.ServerError(
            "Server (%s) is not a valid candidate slave "
            "due to the following reason(s): (%s)." %
            (slave.uuid, master_issues)
            )

    if slave.status not in allowed_status:
        raise _errors.ServerError("Server (%s) is faulty." % (slave_id, ))

    _events.trigger_within_procedure(WAIT_SLAVE_FAIL, group_id, str(slave.uuid))

@_events.on_event(WAIT_SLAVE_FAIL)
def _wait_slave_fail(group_id, slave_uuid):
    """Wait until a slave processes its backlog.
    """
    slave = _server.MySQLServer.fetch(_uuid.UUID(slave_uuid))
    slave.connect()

    try:
        _utils.process_slave_backlog(slave)
    except _errors.DatabaseError as error:
        _LOGGER.warning(
            "Error (%s) trying to process transactions in the relay log "
            "for candidate (%s).", slave, error
        )

    _events.trigger_within_procedure(CHANGE_TO_CANDIDATE, group_id, slave_uuid)

@_events.on_event(BLOCK_WRITE_DEMOTE)
def _block_write_demote(group_id, update_only):
    """Block and disable write access to the current master.
    """
    group = _server.Group.fetch(group_id)
    if not group:
        raise _errors.GroupError("Group (%s) does not exist." % (group_id, ))

    if not group.master:
        raise _errors.GroupError("Group (%s) does not have a master." %
                                 (group_id, ))

    master = _server.MySQLServer.fetch(group.master)
    assert(master.status in \
        (_server.MySQLServer.PRIMARY, _server.MySQLServer.FAULTY)
    )

    if master.status == _server.MySQLServer.PRIMARY:
        master.connect()
        master.mode = _server.MySQLServer.READ_ONLY
        master.status = _server.MySQLServer.SECONDARY
        _utils.set_read_only(master, True)

        if not update_only:
            _events.trigger_within_procedure(
                WAIT_SLAVES_DEMOTE, group_id, str(master.uuid)
            )

    _set_group_master_replication(group, None, update_only)

@_events.on_event(WAIT_SLAVES_DEMOTE)
def _wait_slaves_demote(group_id, master_uuid):
    """Synchronize slaves with master.
    """
    master = _server.MySQLServer.fetch(_uuid.UUID(master_uuid))
    master.connect()

    # Synchronize slaves.
    _do_wait_slaves_catch(group_id, master)

    # Stop replication threads at all slaves.
    group = _server.Group.fetch(group_id)
    for server in group.servers():
        try:
            server.connect()
            _utils.stop_slave(server)
        except _errors.DatabaseError as error:
            _LOGGER.debug(
                "Error waiting for slave (%s) to stop.", server.uuid,
                exc_info=error
            )

def _set_group_master_replication(group, server_id, update_only=False):
    """Set the master for the given group and also reset the
    replication with the other group masters. Any change of master
    for a group will be initiated through this method. The method also
    takes care of resetting the master and the slaves that are registered
    with this group to connect with the new master.

    The idea is that operations like switchover, failover, promote all are
    finally master changing operations. Hence the right place to handle
    these operations is at the place where the master is being changed.

    The following operations need to be done

    - Stop the slave on the old master
    - Stop the slaves replicating from the old master
    - Start the slave on the new master
    - Start the slaves with the new master

    :param group: The group whose master needs to be changed.
    :param server_id: The server id of the server that is becoming the master.
    :param update_only: Only update the state store and skip provisioning.
    """
    # Set the new master if update-only is true.
    if update_only:
        group.master = server_id
        return

    try:
        # Otherwise, stop the slave running on the current master
        if group.master_group_id is not None and group.master is not None:
            _group_replication.stop_group_slave(
                group.master_group_id, group.group_id, False
            )
        # Stop the Groups replicating from the current group.
        _group_replication.stop_group_slaves(group.group_id)
    except (_errors.GroupError, _errors.DatabaseError) as error:
        _LOGGER.error("Error accessing groups related to (%s).",
                      group.group_id)

    # Set the new master
    group.master = server_id

    try:
        # If the master is not None setup the master and the slaves.
        if group.master is not None:
            # Start the slave groups for this group.
            _group_replication.start_group_slaves(group.group_id)
            if group.master_group_id is not None:
                # Start the slave on this group
                _group_replication.setup_group_replication(
                    group.master_group_id, group.group_id
                )
    except (_errors.GroupError, _errors.DatabaseError) as error:
        _LOGGER.error("Error accessing groups related to (%s).",
                      group.group_id)
